import asyncio
from concurrent.futures import ProcessPoolExecutor, ThreadPoolExecutor
from fractions import Fraction
from typing import TYPE_CHECKING, Any, Awaitable, Dict, Iterable, List, Sequence, Tuple, TypeAlias
from warnings import warn

import eth_abi.abi
from eth_typing import ChecksumAddress
from eth_utils.address import to_checksum_address
from scipy.optimize import minimize_scalar
from web3 import Web3

from ..baseclasses import (
    BaseArbitrage,
    BaseLiquidityPool,
    BasePoolState,
    UniswapSimulationResult,
)
from ..config import get_web3
from ..constants import MAX_UINT256
from ..curve.curve_stableswap_dataclasses import CurveStableswapPoolState
from ..curve.curve_stableswap_liquidity_pool import CurveStableswapPool
from ..erc20_token import Erc20Token
from ..exceptions import ArbitrageError, EVMRevertError, LiquidityPoolError, ZeroLiquidityError
from ..logging import logger
from ..subscription_mixins import Publisher, Subscriber
from ..uniswap.v2_dataclasses import UniswapV2PoolSimulationResult, UniswapV2PoolState
from ..uniswap.v2_liquidity_pool import LiquidityPool
from ..uniswap.v3_dataclasses import UniswapV3PoolSimulationResult, UniswapV3PoolState
from ..uniswap.v3_libraries import TickMath
from ..uniswap.v3_liquidity_pool import V3LiquidityPool
from .arbitrage_dataclasses import (
    ArbitrageCalculationResult,
    CurveStableSwapPoolSwapAmounts,
    CurveStableSwapPoolVector,
    UniswapPoolSwapVector,
    UniswapV2PoolSwapAmounts,
    UniswapV3PoolSwapAmounts,
)

SwapAmount: TypeAlias = (
    CurveStableSwapPoolSwapAmounts | UniswapV2PoolSwapAmounts | UniswapV3PoolSwapAmounts
)


# Default discount applied to amount received.
# This masks small differences in get_dy() vs exchange().
CURVE_V1_DEFAULT_DISCOUNT_FACTOR = 0.9999


class UniswapCurveCycle(Subscriber, BaseArbitrage):
    def __init__(
        self,
        input_token: Erc20Token,
        swap_pools: Iterable[BaseLiquidityPool],
        id: str,
        max_input: int | None = None,
    ):
        if any(
            [
                not isinstance(pool, (CurveStableswapPool, LiquidityPool, V3LiquidityPool))
                for pool in swap_pools
            ]
        ):
            raise ValueError("Must provide only Curve StableSwap or Uniswap liquidity pools.")

        self.swap_pools = tuple(swap_pools)
        self.name = " → ".join([pool.name for pool in self.swap_pools])

        self.pool_states: Dict[ChecksumAddress, BasePoolState] = {}
        self._update_pool_states(self.swap_pools)
        self.curve_discount_factor = CURVE_V1_DEFAULT_DISCOUNT_FACTOR

        for pool in swap_pools:
            pool.subscribe(self)

        self.id = id
        self.input_token = input_token

        if max_input is None:
            warn("No maximum input provided, setting to 100 WETH")
            max_input = 100 * 10**18
        self.max_input = max_input

        self.gas_estimate: int

        # Set up pre-determined "swap vectors", which allows the helper
        # to identify the tokens and direction of each swap along the path
        _swap_vectors: List[CurveStableSwapPoolVector | UniswapPoolSwapVector] = []
        for i, pool in enumerate(self.swap_pools):
            match pool:
                case LiquidityPool() | V3LiquidityPool():
                    if i == 0:
                        if self.input_token == pool.token0:
                            token_in = pool.token0
                            token_out = pool.token1
                            zero_for_one = True
                        elif self.input_token == pool.token1:
                            token_in = pool.token1
                            token_out = pool.token0
                            zero_for_one = False
                        else:
                            raise ValueError("Input token could not be identified!")
                    else:
                        if token_out == pool.token0:
                            token_in = pool.token0
                            token_out = pool.token1
                            zero_for_one = True
                        elif token_out == pool.token1:
                            token_in = pool.token1
                            token_out = pool.token0
                            zero_for_one = False
                        else:
                            raise ValueError("Input token could not be identified!")

                    _swap_vectors.append(
                        UniswapPoolSwapVector(
                            token_in=token_in,
                            token_out=token_out,
                            zero_for_one=zero_for_one,
                        )
                    )

                case CurveStableswapPool():
                    # A Curve pool may have 3 or more tokens, so instead of a binary
                    # token0/token1 choice, determine the forward token by comparing
                    # current and next pool
                    if i != 1:
                        raise ValueError(
                            f"Not implemented for Curve pools at position != 1, {i=}, {pool=}, {self.id=}"
                        )
                    token_in = token_out
                    next_pool = self.swap_pools[i + 1]
                    shared_tokens = list(
                        set(pool.tokens).intersection(next_pool.tokens),
                    )
                    assert len(shared_tokens) > 0, f"this: {pool.tokens}, next: {next_pool.tokens}"

                    # @dev
                    # This assumes the first shared token is the correct one to continue
                    token_out = shared_tokens[0]

                    _swap_vectors.append(
                        CurveStableSwapPoolVector(token_in=token_in, token_out=token_out)
                    )

                case _:
                    raise ValueError("Pool type could not be identified")

        self._swap_vectors = tuple(_swap_vectors)

    def __getstate__(self) -> Dict[str, Any]:
        # Remove objects that cannot be pickled and are unnecessary to perform
        # the calculation
        dropped_attributes = (
            "_subscribers",
            "gas_estimate",
        )

        return {key: value for key, value in self.__dict__.items() if key not in dropped_attributes}

    def __str__(self) -> str:
        return self.name

    def _sort_overrides(
        self,
        overrides: Sequence[
            Tuple[
                BaseLiquidityPool,
                BasePoolState | UniswapSimulationResult,
            ]
        ]
        | None,
    ) -> Dict[ChecksumAddress, BasePoolState]:
        """
        Validate the overrides, extract and insert the resulting pool states
        into a dictionary.
        """

        if overrides is None:
            return {}

        sorted_overrides: Dict[ChecksumAddress, BasePoolState] = {}

        for pool, override in overrides:
            if isinstance(
                override,
                (
                    CurveStableswapPoolState,
                    UniswapV2PoolState,
                    UniswapV3PoolState,
                ),
            ):
                logger.debug(f"Applying override {override} to {pool}")
                sorted_overrides[pool.address] = override
            elif isinstance(
                override,
                (
                    # CurveStableswapPoolSimulationResult, <----- todo
                    UniswapV2PoolSimulationResult,
                    UniswapV3PoolSimulationResult,
                ),
            ):
                logger.debug(f"Applying override {override.future_state} to {pool}")
                sorted_overrides[pool.address] = override.future_state
            else:
                raise ValueError(f"Override for {pool} has unsupported type {type(override)}")

        return sorted_overrides

    def _build_amounts_out(
        self,
        token_in: Erc20Token,
        token_in_quantity: int,
        pool_state_overrides: Dict[ChecksumAddress, BasePoolState] | None = None,
        block_number: int | None = None,
    ) -> List[SwapAmount]:
        """
        Generate human-readable inputs for a complete swap along the arbitrage
        path, starting with `token_in_quantity` amount of `token_in`.
        """

        if pool_state_overrides is None:
            pool_state_overrides = {}

        pools_amounts_out: List[SwapAmount] = []

        _token_in_quantity: int = 0
        _token_out_quantity: int = 0

        for i, (pool, swap_vector) in enumerate(zip(self.swap_pools, self._swap_vectors)):
            match pool:
                case LiquidityPool() | V3LiquidityPool():
                    assert isinstance(swap_vector, UniswapPoolSwapVector)
                    token_in = swap_vector.token_in
                    token_out = swap_vector.token_out
                    zero_for_one = swap_vector.zero_for_one
                case CurveStableswapPool():
                    assert isinstance(swap_vector, CurveStableSwapPoolVector)
                    token_in = swap_vector.token_in
                    token_out = swap_vector.token_out

            if i == 0:
                _token_in_quantity = token_in_quantity
            else:
                _token_in_quantity = _token_out_quantity

            try:
                match pool:
                    case LiquidityPool():
                        pool_state_override = pool_state_overrides.get(pool.address)
                        if TYPE_CHECKING:
                            assert pool_state_override is None or isinstance(
                                pool_state_override,
                                UniswapV2PoolState,
                            )
                        _token_out_quantity = pool.calculate_tokens_out_from_tokens_in(
                            token_in=token_in,
                            token_in_quantity=_token_in_quantity,
                            override_state=pool_state_override,
                        )
                    case V3LiquidityPool():
                        pool_state_override = pool_state_overrides.get(pool.address)
                        if TYPE_CHECKING:
                            assert pool_state_override is None or isinstance(
                                pool_state_override,
                                UniswapV3PoolState,
                            )
                        _token_out_quantity = pool.calculate_tokens_out_from_tokens_in(
                            token_in=token_in,
                            token_in_quantity=_token_in_quantity,
                            override_state=pool_state_override,
                        )
                    case CurveStableswapPool():
                        pool_state_override = pool_state_overrides.get(pool.address)
                        if TYPE_CHECKING:
                            assert pool_state_override is None or isinstance(
                                pool_state_override,
                                CurveStableswapPoolState,
                            )
                        _token_out_quantity = int(
                            self.curve_discount_factor
                            * pool.calculate_tokens_out_from_tokens_in(
                                token_in=token_in,
                                token_out=token_out,
                                token_in_quantity=_token_in_quantity,
                                override_state=pool_state_override,
                                block_identifier=block_number,
                            )
                        )
                    case _:
                        raise ValueError(f"Could not determine pool type for {pool}")
            except LiquidityPoolError as e:
                raise ArbitrageError(f"(calculate_tokens_out_from_tokens_in): {e}")
            else:
                if _token_out_quantity == 0:
                    raise ArbitrageError(f"Zero-output swap through pool {pool} @ {pool.address}")

            match pool:
                case LiquidityPool():
                    pools_amounts_out.append(
                        UniswapV2PoolSwapAmounts(
                            amounts=(0, _token_out_quantity)
                            if zero_for_one
                            else (_token_out_quantity, 0),
                        )
                    )
                case V3LiquidityPool():
                    pools_amounts_out.append(
                        UniswapV3PoolSwapAmounts(
                            amount_specified=_token_in_quantity,
                            zero_for_one=zero_for_one,
                            sqrt_price_limit_x96=TickMath.MIN_SQRT_RATIO + 1
                            if zero_for_one
                            else TickMath.MAX_SQRT_RATIO - 1,
                        )
                    )
                case CurveStableswapPool():
                    pools_amounts_out.append(
                        CurveStableSwapPoolSwapAmounts(
                            token_in=token_in,
                            token_in_index=pool.tokens.index(token_in),
                            token_out=token_out,
                            token_out_index=pool.tokens.index(token_out),
                            amount_in=_token_in_quantity,
                            min_amount_out=_token_out_quantity,
                            underlying=True
                            if (
                                pool.is_metapool
                                and (
                                    token_in in pool.tokens_underlying
                                    or token_out in pool.tokens_underlying
                                )
                            )
                            else False,
                        )
                    )
                case _:
                    raise ValueError(
                        f"Could not identify Uniswap version for pool: {self.swap_pools[i]}"
                    )

        return pools_amounts_out

    def _update_pool_states(self, pools: Iterable[BaseLiquidityPool]) -> None:
        """
        Update `self.pool_states` with state values from the `pools` iterable
        """
        self.pool_states.update({pool.address: pool.state for pool in pools})

    def _pre_calculation_check(
        self,
        override_state: Sequence[
            Tuple[
                BaseLiquidityPool,
                BasePoolState | UniswapSimulationResult,
            ]
        ]
        | None = None,
    ) -> None:
        state_overrides = self._sort_overrides(override_state)

        # A scalar value representing the net amount of 1 input token across
        # the complete path (including fees).
        # profit_factor > 1.0 indicates a profitable trade.
        profit_factor: float = 1.0

        # Check each pool for liquidity in the direction of the trade and account for its current price and fee
        # The prices are absolute (not decimal-corrected) since the decimals for intermediate tokens cancel out
        # e.g. for a WETH -> USDC -> USDT -> WETH arbitrage,
        # profit factor: [input: WETH] -> [pool0: USDC/WETH] * [pool1: USDT/USDC] * [pool2: WETH/USDT] == [output: WETH]
        for pool, vector in zip(self.swap_pools, self._swap_vectors):
            pool_state = state_overrides.get(pool.address) or pool.state

            match pool:
                case LiquidityPool():
                    if TYPE_CHECKING:
                        assert isinstance(pool_state, UniswapV2PoolState)
                        assert isinstance(vector, UniswapPoolSwapVector)

                    if pool_state.reserves_token0 == 0 or pool_state.reserves_token1 == 0:
                        raise ZeroLiquidityError(f"V2 pool {pool.address} has no liquidity")

                    if pool_state.reserves_token1 == 1 and vector.zero_for_one:
                        raise ZeroLiquidityError(
                            f"V2 pool {pool.address} has no liquidity for a 0 -> 1 swap"
                        )
                    elif pool_state.reserves_token0 == 1 and not vector.zero_for_one:
                        raise ZeroLiquidityError(
                            f"V2 pool {pool.address} has no liquidity for a 1 -> 0 swap"
                        )

                    price = pool_state.reserves_token1 / pool_state.reserves_token0
                    fee = pool.fee_token0 if vector.zero_for_one else pool.fee_token1
                    profit_factor *= (price if vector.zero_for_one else 1 / price) * (
                        (fee.denominator - fee.numerator) / fee.denominator
                    )

                case V3LiquidityPool():
                    if TYPE_CHECKING:
                        assert isinstance(pool_state, UniswapV3PoolState)
                        assert isinstance(vector, UniswapPoolSwapVector)

                    if pool_state.sqrt_price_x96 == 0:
                        raise ZeroLiquidityError(
                            f"V3 pool {pool.address} has no liquidity (not initialized)"
                        )

                    if pool_state.tick_bitmap == {}:
                        raise ZeroLiquidityError(
                            f"V3 pool {pool.address} has no liquidity (empty bitmap)"
                        )

                    if pool_state.liquidity == 0:
                        # Check if the swap is 0 -> 1 and cannot swap any more
                        # token0 for token1
                        if (
                            pool_state.sqrt_price_x96 == TickMath.MIN_SQRT_RATIO + 1
                            and vector.zero_for_one
                        ):
                            raise ZeroLiquidityError(
                                f"V3 pool {pool.address} has no liquidity for a 0 -> 1 swap"
                            )
                        # Check if the swap is 1 -> 0 (zeroForOne=False) and
                        # cannot swap any more token1 for token0
                        elif (
                            pool_state.sqrt_price_x96 == TickMath.MAX_SQRT_RATIO - 1
                            and not vector.zero_for_one
                        ):
                            raise ZeroLiquidityError(
                                f"V3 pool {pool.address} has no liquidity for a 1 -> 0 swap"
                            )

                    price = pool_state.sqrt_price_x96**2 / (2**192)
                    # V3 fees are integer values representing hundredths of a bip (0.0001)
                    # e.g. fee=3000 represents 0.3%
                    fee = Fraction(pool._fee, 1000000)
                    profit_factor *= (price if vector.zero_for_one else 1 / price) * (
                        (fee.denominator - fee.numerator) / fee.denominator
                    )

                case CurveStableswapPool():
                    price = 1.0 * (10**vector.token_out.decimals) / (10**vector.token_in.decimals)
                    fee = Fraction(pool.fee, pool.FEE_DENOMINATOR)
                    profit_factor *= price * ((fee.denominator - fee.numerator) / fee.denominator)

                case _:
                    raise ValueError("Could not identify pool")

            # print(f"{profit_factor=}")

        if profit_factor < 1.0:
            raise ArbitrageError(
                f"No profitable arbitrage at current prices. Profit factor: {profit_factor}"
            )

    def _calculate(
        self,
        override_state: Sequence[
            Tuple[
                BaseLiquidityPool,
                BasePoolState | UniswapSimulationResult,
            ]
        ]
        | None = None,
        block_number: int | None = None,
    ) -> ArbitrageCalculationResult:
        state_overrides = self._sort_overrides(override_state)

        # bound the amount to be swapped
        bounds = (
            1.0,
            max(2.0, float(self.max_input)),
        )

        # bracket the initial guess for the algo
        bracket_amount = self.max_input
        bracket = (
            0.45 * bracket_amount,
            0.50 * bracket_amount,
            0.55 * bracket_amount,
        )

        def arb_profit(x: float) -> float:
            token_in_quantity = int(x)  # round the input down
            token_out_quantity: int = 0

            for i, (pool, swap_vector) in enumerate(zip(self.swap_pools, self._swap_vectors)):
                pool_override = state_overrides.get(pool.address)

                try:
                    match pool:
                        case LiquidityPool():
                            if TYPE_CHECKING:
                                assert isinstance(pool_override, UniswapV2PoolState)
                            token_out_quantity = pool.calculate_tokens_out_from_tokens_in(
                                token_in=swap_vector.token_in,
                                token_in_quantity=token_in_quantity
                                if i == 0
                                else token_out_quantity,
                                override_state=pool_override,
                            )
                        case V3LiquidityPool():
                            if TYPE_CHECKING:
                                assert isinstance(pool_override, UniswapV3PoolState)
                            token_out_quantity = pool.calculate_tokens_out_from_tokens_in(
                                token_in=swap_vector.token_in,
                                token_in_quantity=token_in_quantity
                                if i == 0
                                else token_out_quantity,
                                override_state=pool_override,
                            )
                        case CurveStableswapPool():
                            if TYPE_CHECKING:
                                assert isinstance(pool_override, CurveStableswapPoolState)
                            token_out_quantity = int(
                                self.curve_discount_factor
                                * pool.calculate_tokens_out_from_tokens_in(
                                    token_in=swap_vector.token_in,
                                    token_in_quantity=(
                                        token_in_quantity if i == 0 else token_out_quantity
                                    ),
                                    token_out=swap_vector.token_out,
                                    override_state=pool_override,
                                    block_identifier=block_number,
                                )
                            )
                        case _:
                            raise ValueError

                except (EVMRevertError, LiquidityPoolError):
                    # The optimizer might send invalid amounts into the swap
                    # calculation during iteration. We don't want it to stop,
                    # so catch the exception and pretend the swap results in
                    # token_out_quantity = 0.
                    token_out_quantity = 0
                    break

            # minimize_scalar requires the function to have a minimum value
            # for the solver to settle on an optimum input, so return the
            # negated profit
            return -float(token_out_quantity - token_in_quantity)

        opt = minimize_scalar(
            fun=arb_profit,
            method="bounded",
            bounds=bounds,
            bracket=bracket,
            options={"xatol": 1.0},
        )

        # Negate the result to convert to a sensible value (positive profit)
        best_profit = -int(opt.fun)
        swap_amount = int(opt.x)

        try:
            best_amounts = self._build_amounts_out(
                token_in=self.input_token,
                token_in_quantity=swap_amount,
                pool_state_overrides=state_overrides,
                block_number=block_number,
            )
        # except (EVMRevertError, LiquidityPoolError) as e:
        except ArbitrageError as e:
            # Simulated EVM reverts inside the ported `swap` function were
            # ignored to execute the optimizer to completion. Now the optimal
            # value should be tested and raise an exception if it would
            # generate a bad payload that will revert
            raise ArbitrageError(f"No possible arbitrage: {e}") from None
        except Exception as e:
            raise ArbitrageError(f"No possible arbitrage: {e}") from e

        return ArbitrageCalculationResult(
            id=self.id,
            input_token=self.input_token,
            profit_token=self.input_token,
            input_amount=swap_amount,
            profit_amount=best_profit,
            swap_amounts=best_amounts,
        )

    def calculate(
        self,
        override_state: Sequence[
            Tuple[
                BaseLiquidityPool,
                BasePoolState | UniswapSimulationResult,
            ]
        ]
        | None = None,
    ) -> ArbitrageCalculationResult:
        """
        Calculate the optimum arbitrage input and intermediate swap values for the current pool states.
        """

        self._pre_calculation_check(override_state)

        return self._calculate(override_state=override_state)

    async def calculate_with_pool(
        self,
        executor: ProcessPoolExecutor | ThreadPoolExecutor,
        override_state: Sequence[
            Tuple[
                BaseLiquidityPool,
                BasePoolState | UniswapSimulationResult,
            ]
        ]
        | None = None,
    ) -> Awaitable[Any]:
        """
        Wrap the arbitrage calculation into an asyncio future using the
        specified executor.

        Arguments
        ---------
        executor : Executor
            An executor (from `concurrent.futures`) to process the calculation
            work. Both `ThreadPoolExecutor` and `ProcessPoolExecutor` are
            supported, but `ProcessPoolExecutor` is recommended.
        override_state : StateOverrideTypes, optional
            An sequence of tuples, representing an ordered pair of helper
            objects for Uniswap V2 / V3 pools and their overridden states.

        Returns
        -------
        A future which returns a `ArbitrageCalculationResult` (or exception)
        when awaited.

        Notes
        -----
        This is an async function that must be called with the `await` keyword.
        """

        self._pre_calculation_check(override_state)

        if any(
            [pool._sparse_bitmap for pool in self.swap_pools if isinstance(pool, V3LiquidityPool)]
        ):
            raise ValueError(
                f"Cannot calculate {self} with executor. One or more V3 pools has a sparse bitmap."
            )

        curve_pool = self.swap_pools[1]
        curve_swap_vector = self._swap_vectors[1]

        if TYPE_CHECKING:
            assert isinstance(curve_pool, CurveStableswapPool)
            assert isinstance(curve_swap_vector, CurveStableSwapPoolVector)

        block_number = get_web3().eth.get_block_number()

        # Some Curve pools utilize on-chain lookups in their calc, so do a simple pre-calc to
        # cache those values for a given block since the pool will be disconnected once sent
        # into the process pool, e.g. it will have no web3 object for communication with the chain
        curve_pool.calculate_tokens_out_from_tokens_in(
            token_in=curve_swap_vector.token_in,
            token_in_quantity=1,
            token_out=curve_swap_vector.token_out,
            block_identifier=block_number,
        )

        return asyncio.get_running_loop().run_in_executor(
            executor,
            self._calculate,
            override_state,
            block_number,
        )

    def generate_payloads(
        self,
        from_address: ChecksumAddress | str,
        swap_amount: int,
        pool_swap_amounts: Sequence[
            CurveStableSwapPoolSwapAmounts | UniswapV2PoolSwapAmounts | UniswapV3PoolSwapAmounts
        ],
        infinite_approval: bool = False,
    ) -> List[Tuple[ChecksumAddress, bytes, int]]:
        """
        TBD
        """

        from_address = to_checksum_address(from_address)

        # Abandon empty inputs.
        # @dev this looks like a useful place for a ValueError, but threaded
        # clients may execute a pool update for a swap pool before the call to
        # generate payloads is processed. Abandon the call in this case and
        # raise a generic non-fatal exception.
        if not pool_swap_amounts:
            raise ArbitrageError("Pool amounts empty, abandoning payload generation.")

        payloads = []
        msg_value: int = 0  # This arbitrage does not require a `msg.value` payment

        first_pool = self.swap_pools[0]
        last_pool = self.swap_pools[-1]

        try:
            if isinstance(first_pool, LiquidityPool):
                # Special case: If first pool is type V2, input token must be
                # transferred prior to the swap
                logger.debug(f"PAYLOAD: transferring {swap_amount} WETH to V2 pool {first_pool}")
                payloads.append(
                    (
                        # address
                        self.input_token.address,
                        # bytes calldata
                        Web3.keccak(text="transfer(address,uint256)")[:4]
                        + eth_abi.abi.encode(
                            types=(
                                "address",
                                "uint256",
                            ),
                            args=(
                                first_pool.address,
                                swap_amount,
                            ),
                        ),
                        msg_value,
                    )
                )

            for i, (swap_pool, _swap_amounts) in enumerate(zip(self.swap_pools, pool_swap_amounts)):
                if swap_pool is last_pool:
                    next_pool = None
                else:
                    next_pool = self.swap_pools[i + 1]

                if next_pool is not None:
                    # V2 pools require a pre-swap transfer, so the contract
                    # does not have to perform intermediate custody and the
                    # swap can send the tokens directly to the next pool
                    if isinstance(next_pool, LiquidityPool):
                        swap_destination_address = next_pool.address
                    # V3 pools cannot accept a pre-swap transfer, so the contract
                    # must maintain custody prior to a swap
                    elif isinstance(next_pool, V3LiquidityPool):
                        swap_destination_address = from_address
                    # Curve V1 pools execute a transferFrom on behalf of msg.sender, so the contract
                    # must maintain custody prior to a swap
                    elif isinstance(next_pool, CurveStableswapPool):
                        swap_destination_address = from_address
                else:
                    # Set the destination address for the last swap to the
                    # sending address
                    swap_destination_address = from_address

                if isinstance(swap_pool, LiquidityPool):
                    if TYPE_CHECKING:
                        assert isinstance(_swap_amounts, UniswapV2PoolSwapAmounts)
                    if _swap_amounts.amounts[0] == 0:
                        _token_in = swap_pool.token0
                        _token_out = swap_pool.token1
                        _amount_out = _swap_amounts.amounts[1]
                    else:
                        _token_in = swap_pool.token1
                        _token_out = swap_pool.token0
                        _amount_out = _swap_amounts.amounts[0]
                    logger.debug(f"PAYLOAD: building V2 swap at pool {i}")
                    logger.debug(f"PAYLOAD: pool address {swap_pool.address}")
                    logger.debug(f"PAYLOAD: swap {_token_in} -> {_amount_out} {_token_out} ")
                    logger.debug(f"PAYLOAD: destination address {swap_destination_address}")
                    payloads.append(
                        (
                            # address
                            swap_pool.address,
                            # bytes calldata
                            Web3.keccak(text="swap(uint256,uint256,address,bytes)")[:4]
                            + eth_abi.abi.encode(
                                types=(
                                    "uint256",
                                    "uint256",
                                    "address",
                                    "bytes",
                                ),
                                args=(
                                    *_swap_amounts.amounts,
                                    swap_destination_address,
                                    b"",
                                ),
                            ),
                            msg_value,
                        )
                    )
                elif isinstance(swap_pool, V3LiquidityPool):
                    if TYPE_CHECKING:
                        assert isinstance(_swap_amounts, UniswapV3PoolSwapAmounts)
                    if _swap_amounts.zero_for_one:
                        _token_in = swap_pool.token0
                        _token_out = swap_pool.token1
                    else:
                        _token_in = swap_pool.token1
                        _token_out = swap_pool.token0
                    logger.debug(f"PAYLOAD: building V3 swap at pool {i}")
                    logger.debug(f"PAYLOAD: pool address {swap_pool.address}")
                    logger.debug(f"PAYLOAD: token in  = {_token_in}")
                    logger.debug(f"PAYLOAD: token out = {_token_out}")
                    logger.debug(f"PAYLOAD: swap amounts {_swap_amounts}")
                    logger.debug(f"PAYLOAD: destination address {swap_destination_address}")
                    payloads.append(
                        (
                            # address
                            swap_pool.address,
                            # bytes calldata
                            Web3.keccak(text="swap(address,bool,int256,uint160,bytes)")[:4]
                            + eth_abi.abi.encode(
                                types=(
                                    "address",
                                    "bool",
                                    "int256",
                                    "uint160",
                                    "bytes",
                                ),
                                args=(
                                    swap_destination_address,
                                    _swap_amounts.zero_for_one,
                                    _swap_amounts.amount_specified,
                                    _swap_amounts.sqrt_price_limit_x96,
                                    b"",
                                ),
                            ),
                            msg_value,
                        )
                    )
                elif isinstance(swap_pool, CurveStableswapPool):
                    if TYPE_CHECKING:
                        assert isinstance(_swap_amounts, CurveStableSwapPoolSwapAmounts)
                    logger.debug(f"PAYLOAD: building Curve swap at pool {i}")
                    logger.debug(f"PAYLOAD: pool address {swap_pool.address}")
                    logger.debug(f"PAYLOAD: swap amounts {_swap_amounts}")
                    logger.debug(f"PAYLOAD: token in  = {_swap_amounts.token_in}")
                    logger.debug(f"PAYLOAD: token out = {_swap_amounts.token_out}")
                    logger.debug(f"PAYLOAD: destination address {swap_destination_address}")

                    current_approval = _swap_amounts.token_in.get_approval(
                        from_address, swap_pool.address
                    )
                    amount_to_approve: int | None = None
                    if infinite_approval is True and current_approval != MAX_UINT256:
                        amount_to_approve = MAX_UINT256
                    elif infinite_approval is False and current_approval < _swap_amounts.amount_in:
                        amount_to_approve = _swap_amounts.amount_in

                    if amount_to_approve is not None:
                        logger.debug(
                            f"PAYLOAD: approve {amount_to_approve} {_swap_amounts.token_in} by {swap_pool} {swap_destination_address}"
                        )
                        payloads.append(
                            (
                                # address
                                _swap_amounts.token_in.address,
                                # bytes calldata
                                Web3.keccak(text="approve(address,uint256)")[:4]
                                + eth_abi.abi.encode(
                                    types=["address", "uint256"],
                                    args=[swap_pool.address, amount_to_approve],
                                ),
                                msg_value,
                            )
                        )

                    logger.debug(
                        f"PAYLOAD: exchange {_swap_amounts.amount_in} {_swap_amounts.token_in_index}->{_swap_amounts.token_out_index}, min out = {_swap_amounts.min_amount_out}"
                    )
                    if _swap_amounts.underlying:
                        payloads.append(
                            (
                                # address
                                swap_pool.address,
                                # bytes calldata
                                Web3.keccak(
                                    text="exchange_underlying(int128,int128,uint256,uint256)"
                                )[:4]
                                + eth_abi.abi.encode(
                                    types=["int128", "int128", "uint256", "uint256"],
                                    args=[
                                        _swap_amounts.token_in_index,
                                        _swap_amounts.token_out_index,
                                        _swap_amounts.amount_in,
                                        _swap_amounts.min_amount_out,
                                    ],
                                ),
                                msg_value,
                            )
                        )
                    else:
                        payloads.append(
                            (
                                # address
                                swap_pool.address,
                                # bytes calldata
                                Web3.keccak(text="exchange(int128,int128,uint256,uint256)")[:4]
                                + eth_abi.abi.encode(
                                    types=["int128", "int128", "uint256", "uint256"],
                                    args=[
                                        _swap_amounts.token_in_index,
                                        _swap_amounts.token_out_index,
                                        _swap_amounts.amount_in,
                                        _swap_amounts.min_amount_out,
                                    ],
                                ),
                                msg_value,
                            )
                        )
                    if next_pool is not None and isinstance(next_pool, LiquidityPool):
                        # V2 pools require a pre-swap transfer, so the contract
                        # does not have to perform intermediate custody and the
                        # swap can send the tokens directly to the next pool
                        logger.debug(
                            f"PAYLOAD: transferring {_swap_amounts.min_amount_out} {_swap_amounts.token_out} to V2 pool {next_pool}"
                        )
                        payloads.append(
                            (
                                # address
                                _swap_amounts.token_out.address,
                                # bytes calldata
                                Web3.keccak(text="transfer(address,uint256)")[:4]
                                + eth_abi.abi.encode(
                                    types=(
                                        "address",
                                        "uint256",
                                    ),
                                    args=(
                                        next_pool.address,
                                        _swap_amounts.min_amount_out,
                                    ),
                                ),
                                msg_value,
                            )
                        )

                else:
                    raise ValueError(
                        f"Could not identify pool: {swap_pool}, type={type(swap_pool)}"
                    )
        except Exception as e:
            logger.exception("generate_payloads catch-all")
            raise ArbitrageError(f"generate_payloads (catch-all)): {e}") from e

        return payloads

    def notify(self, publisher: Publisher) -> None:
        # On receipt of a notification from a publishing pool, update the pool state
        if isinstance(publisher, (LiquidityPool, V3LiquidityPool)):
            self._update_pool_states((publisher,))



import itertools
from typing import TYPE_CHECKING, Iterable, List, Sequence

import eth_abi.packed
from eth_typing import ChecksumAddress
from eth_utils.address import to_checksum_address
from hexbytes import HexBytes
from web3 import Web3

if TYPE_CHECKING:
    from .managers import UniswapV2LiquidityPoolManager
    from .v2_liquidity_pool import LiquidityPool


def generate_v2_pool_address(
    token_addresses: Sequence[ChecksumAddress | str],
    factory_address: ChecksumAddress | str,
    init_hash: str,
) -> ChecksumAddress:
    """
    Generate the deterministic pool address from the token addresses.

    Adapted from https://github.com/Uniswap/universal-router/blob/deployed-commit/contracts/modules/uniswap/v2/UniswapV2Library.sol
    """

    token_addresses = sorted([address.lower() for address in token_addresses])

    return to_checksum_address(
        Web3.keccak(
            HexBytes(0xFF)
            + HexBytes(factory_address)
            + Web3.keccak(
                eth_abi.packed.encode_packed(
                    ["address", "address"],
                    [*token_addresses],
                )
            )
            + HexBytes(init_hash)
        )[-20:]
    )


def get_v2_pools_from_token_path(
    tx_path: Iterable[ChecksumAddress | str],
    pool_manager: "UniswapV2LiquidityPoolManager",
) -> List["LiquidityPool"]:
    return [
        pool_manager.get_pool(
            token_addresses=token_addresses,
            silent=True,
        )
        for token_addresses in itertools.pairwise(tx_path)
    ]



from decimal import Decimal
from typing import Dict, Tuple

from ...constants import MAX_UINT8
from ...exceptions import BitmapWordUnavailableError, EVMRevertError, MissingTickWordError
from ...logging import logger
from ..v3_dataclasses import UniswapV3BitmapAtWord
from . import bit_math as BitMath


def flipTick(
    tick_bitmap: Dict[int, "UniswapV3BitmapAtWord"],
    tick: int,
    tick_spacing: int,
    update_block: int | None = None,
) -> None:
    if not (tick % tick_spacing == 0):
        raise EVMRevertError("Tick not correctly spaced!")

    word_pos, bit_pos = position(int(Decimal(tick) // tick_spacing))
    logger.debug(f"Flipping {tick=} @ {word_pos=}, {bit_pos=}")

    try:
        mask = 1 << bit_pos
        tick_bitmap[word_pos].bitmap ^= mask
        tick_bitmap[word_pos].block = update_block
    except KeyError:
        raise MissingTickWordError(f"Called flipTick on missing word={word_pos}")
    else:
        logger.debug(f"Flipped {tick=} @ {word_pos=}, {bit_pos=}")


def position(tick: int) -> Tuple[int, int]:
    word_pos: int = tick >> 8
    bit_pos: int = tick % 256
    return (word_pos, bit_pos)


def nextInitializedTickWithinOneWord(
    tick_bitmap: Dict[int, "UniswapV3BitmapAtWord"],
    tick: int,
    tick_spacing: int,
    less_than_or_equal: bool,
) -> Tuple[int, bool]:
    compressed = int(
        Decimal(tick) // tick_spacing
    )  # tick can be negative, use Decimal so floor division rounds to zero instead of negative infinity
    if tick < 0 and tick % tick_spacing != 0:
        compressed -= 1  # round towards negative infinity

    if less_than_or_equal:
        word_pos, bit_pos = position(compressed)

        try:
            bitmap_at_word = tick_bitmap[word_pos].bitmap
        except KeyError:
            raise BitmapWordUnavailableError(f"Bitmap at word {word_pos} unavailable.", word_pos)

        # all the 1s at or to the right of the current bitPos
        mask = 2 * (1 << bit_pos) - 1
        masked = bitmap_at_word & mask

        # if there are no initialized ticks to the right of or at the current tick, return rightmost in the word
        initialized_status = masked != 0
        # overflow/underflow is possible, but prevented externally by limiting both tickSpacing and tick
        next_tick = (
            (compressed - (bit_pos - BitMath.mostSignificantBit(masked))) * tick_spacing
            if initialized_status
            else (compressed - (bit_pos)) * tick_spacing
        )
    else:
        # start from the word of the next tick, since the current tick state doesn't matter
        word_pos, bit_pos = position(compressed + 1)

        try:
            bitmap_at_word = tick_bitmap[word_pos].bitmap
        except KeyError:
            raise BitmapWordUnavailableError(f"Bitmap at word {word_pos} unavailable.", word_pos)

        # all the 1s at or to the left of the bitPos
        mask = ~((1 << bit_pos) - 1)
        masked = bitmap_at_word & mask

        # if there are no initialized ticks to the left of the current tick, return leftmost in the word
        initialized_status = masked != 0
        # overflow/underflow is possible, but prevented externally by limiting both tickSpacing and tick
        next_tick = (
            (compressed + 1 + (BitMath.leastSignificantBit(masked) - bit_pos)) * tick_spacing
            if initialized_status
            else (compressed + 1 + (MAX_UINT8 - bit_pos)) * tick_spacing
        )

    return next_tick, initialized_status



# ruff: noqa: F401
from .arbitrage_dataclasses import ArbitrageCalculationResult
from .flash_borrow_to_lp_swap import FlashBorrowToLpSwap
from .flash_borrow_to_lp_swap_new import FlashBorrowToLpSwapNew
from .flash_borrow_to_lp_swap_with_future import FlashBorrowToLpSwapWithFuture
from .flash_borrow_to_router_swap import FlashBorrowToRouterSwap
from .uniswap_curve_cycle import UniswapCurveCycle
from .uniswap_lp_cycle import UniswapLpCycle



from fractions import Fraction


from ..erc20_token import Erc20Token
from ..exceptions import ZeroSwapError
from ..logging import logger
from .v2_dataclasses import UniswapV2PoolState


class CamelotStablePoolMixin:
    token0: Erc20Token
    token1: Erc20Token
    fee_token0: Fraction
    fee_token1: Fraction
    fee_denominator: int
    reserves_token0: int
    reserves_token1: int
    stable_swap: bool

    def _calculate_tokens_out_from_tokens_in_stable_swap(
        self,
        token_in: "Erc20Token",
        token_in_quantity: int,
        override_state: UniswapV2PoolState | None = None,
    ) -> int:
        """
        Calculates the expected token OUTPUT for a target INPUT at current pool reserves.
        Uses the self.token0 and self.token1 pointers to determine which token is being swapped in
        """

        if override_state is not None:
            logger.debug("Reserve overrides applied:")
            logger.debug(f"token0: {override_state.reserves_token0}")
            logger.debug(f"token1: {override_state.reserves_token1}")

        if token_in_quantity <= 0:
            raise ZeroSwapError("token_in_quantity must be positive")

        precision_multiplier_token0: int = 10**self.token0.decimals
        precision_multiplier_token1: int = 10**self.token1.decimals

        def _k(balance_0: int, balance_1: int) -> int:
            _x: int = balance_0 * 10**18 // precision_multiplier_token0
            _y: int = balance_1 * 10**18 // precision_multiplier_token1
            _a: int = _x * _y // 10**18
            _b: int = (_x * _x // 10**18) + (_y * _y // 10**18)
            return _a * _b // 10**18  # x^3*y+y^3*x >= k

        def _get_y(x_0: int, xy: int, y: int) -> int:
            for _ in range(255):
                y_prev = y
                k = _f(x_0, y)
                if k < xy:
                    dy = (xy - k) * 10**18 // _d(x_0, y)
                    y = y + dy
                else:
                    dy = (k - xy) * 10**18 // _d(x_0, y)
                    y = y - dy

                if y > y_prev:
                    if y - y_prev <= 1:
                        return y
                else:
                    if y_prev - y <= 1:
                        return y

            return y

        def _f(x_0: int, y: int) -> int:
            return (
                x_0 * (y * y // 10**18 * y // 10**18) // 10**18
                + (x_0 * x_0 // 10**18 * x_0 // 10**18) * y // 10**18
            )

        def _d(x_0: int, y: int) -> int:
            return 3 * x_0 * (y * y // 10**18) // 10**18 + (x_0 * x_0 // 10**18 * x_0 // 10**18)

        # fee_percent is stored as a uint16 in the contract, but as a Fraction
        # in the superclass, so it must be converted.
        #
        # e.g. 0.04% fee = Fraction(1,2500) in the helper, fee = 40 in the
        # contract. To convert, multiply the fraction by the `FEE_DENOMINATOR`,
        # so fee_percent = 1/2500 * 100_000 = 40

        fee_percent = (
            self.fee_token0 if token_in is self.token0 else self.fee_token1
        ) * self.fee_denominator

        reserves_token0 = (
            override_state.reserves_token0 if override_state is not None else self.reserves_token0
        )
        reserves_token1 = (
            override_state.reserves_token1 if override_state is not None else self.reserves_token1
        )

        # remove fee from amount received
        token_in_quantity -= token_in_quantity * fee_percent // self.fee_denominator
        xy = _k(reserves_token0, reserves_token1)
        reserves_token0 = reserves_token0 * 10**18 // precision_multiplier_token0
        reserves_token1 = reserves_token1 * 10**18 // precision_multiplier_token1
        reserve_a, reserve_b = (
            (reserves_token0, reserves_token1)
            if token_in is self.token0
            else (reserves_token1, reserves_token0)
        )
        token_in_quantity = (
            token_in_quantity * 10**18 // precision_multiplier_token0
            if token_in is self.token0
            else token_in_quantity * 10**18 // precision_multiplier_token1
        )
        y = reserve_b - _get_y(token_in_quantity + reserve_a, xy, reserve_b)

        return (
            y
            * (
                precision_multiplier_token1
                if token_in is self.token0
                else precision_multiplier_token0
            )
            // 10**18
        )



from ...constants import MIN_UINT160
from ...exceptions import EVMRevertError
from . import full_math as FullMath
from . import unsafe_math as UnsafeMath
from .constants import Q96, Q96_RESOLUTION
from .functions import to_int256, to_uint160


def getAmount0Delta(
    sqrtRatioAX96: int,
    sqrtRatioBX96: int,
    liquidity: int,
    roundUp: bool | None = None,
) -> int:
    # The Solidity function is overloaded with respect to `roundUp`.
    # ref: https://github.com/Uniswap/v3-core/blob/main/contracts/libraries/SqrtPriceMath.sol

    if roundUp is not None:
        if sqrtRatioAX96 > sqrtRatioBX96:
            sqrtRatioAX96, sqrtRatioBX96 = sqrtRatioBX96, sqrtRatioAX96

        numerator1 = liquidity << Q96_RESOLUTION
        numerator2 = sqrtRatioBX96 - sqrtRatioAX96

        if not (sqrtRatioAX96 > 0):
            raise EVMRevertError("require sqrtRatioAX96 > 0")

        return (
            UnsafeMath.divRoundingUp(
                FullMath.mulDivRoundingUp(numerator1, numerator2, sqrtRatioBX96), sqrtRatioAX96
            )
            if roundUp
            else FullMath.mulDiv(numerator1, numerator2, sqrtRatioBX96) // sqrtRatioAX96
        )
    else:
        return to_int256(
            to_int256(-getAmount0Delta(sqrtRatioAX96, sqrtRatioBX96, -liquidity, False))
            if liquidity < 0
            else to_int256(getAmount0Delta(sqrtRatioAX96, sqrtRatioBX96, liquidity, True))
        )


def getAmount1Delta(
    sqrtRatioAX96: int,
    sqrtRatioBX96: int,
    liquidity: int,
    roundUp: bool | None = None,
) -> int:
    # The Solidity function is overloaded with respect to `roundUp`.
    # ref: https://github.com/Uniswap/v3-core/blob/main/contracts/libraries/SqrtPriceMath.sol

    if roundUp is not None:
        if sqrtRatioAX96 > sqrtRatioBX96:
            sqrtRatioAX96, sqrtRatioBX96 = sqrtRatioBX96, sqrtRatioAX96

        return (
            FullMath.mulDivRoundingUp(liquidity, sqrtRatioBX96 - sqrtRatioAX96, Q96)
            if roundUp
            else FullMath.mulDiv(liquidity, sqrtRatioBX96 - sqrtRatioAX96, Q96)
        )
    else:
        return to_int256(
            to_int256(-getAmount1Delta(sqrtRatioAX96, sqrtRatioBX96, -liquidity, False))
            if liquidity < 0
            else to_int256(getAmount1Delta(sqrtRatioAX96, sqrtRatioBX96, liquidity, True))
        )


def getNextSqrtPriceFromAmount0RoundingUp(
    sqrtPX96: int,
    liquidity: int,
    amount: int,
    add: bool,
) -> int:
    # we short circuit amount == 0 because the result is otherwise not guaranteed to equal the input price
    if amount == 0:
        return sqrtPX96

    numerator1 = liquidity << Q96_RESOLUTION

    if add:
        product = amount * sqrtPX96
        if product // amount == sqrtPX96:
            denominator = numerator1 + product
            if denominator >= numerator1:
                # always fits in 160 bits
                return FullMath.mulDivRoundingUp(numerator1, sqrtPX96, denominator)
        return UnsafeMath.divRoundingUp(numerator1, numerator1 // sqrtPX96 + amount)
    else:
        product = amount * sqrtPX96
        # if the product overflows, we know the denominator underflows
        # in addition, we must check that the denominator does not underflow

        if not (product // amount == sqrtPX96 and numerator1 > product):
            raise EVMRevertError("product / amount == sqrtPX96 && numerator1 > product")

        denominator = numerator1 - product
        return to_uint160(FullMath.mulDivRoundingUp(numerator1, sqrtPX96, denominator))


def getNextSqrtPriceFromAmount1RoundingDown(
    sqrtPX96: int,
    liquidity: int,
    amount: int,
    add: bool,
) -> int:
    if add:
        quotient = (
            (amount << Q96_RESOLUTION) // liquidity
            if amount <= 2**160 - 1
            else FullMath.mulDiv(amount, Q96, liquidity)
        )
        return to_uint160(sqrtPX96 + quotient)
    else:
        quotient = (
            UnsafeMath.divRoundingUp(amount << Q96_RESOLUTION, liquidity)
            if amount <= (2**160) - 1
            else FullMath.mulDivRoundingUp(amount, Q96, liquidity)
        )

        if not (sqrtPX96 > quotient):
            raise EVMRevertError("require sqrtPX96 > quotient")

        # always fits 160 bits
        return sqrtPX96 - quotient


def getNextSqrtPriceFromInput(
    sqrtPX96: int,
    liquidity: int,
    amountIn: int,
    zeroForOne: bool,
) -> int:
    if not (sqrtPX96 > MIN_UINT160):
        raise EVMRevertError("sqrtPX96 must be greater than 0")

    if not (liquidity > MIN_UINT160):
        raise EVMRevertError("liquidity must be greater than 0")

    # round to make sure that we don't pass the target price
    return (
        getNextSqrtPriceFromAmount0RoundingUp(sqrtPX96, liquidity, amountIn, True)
        if zeroForOne
        else getNextSqrtPriceFromAmount1RoundingDown(sqrtPX96, liquidity, amountIn, True)
    )


def getNextSqrtPriceFromOutput(
    sqrtPX96: int,
    liquidity: int,
    amountOut: int,
    zeroForOne: bool,
) -> int:
    if not (sqrtPX96 > 0):
        raise EVMRevertError

    if not (liquidity > 0):
        raise EVMRevertError

    # round to make sure that we pass the target price
    return (
        getNextSqrtPriceFromAmount1RoundingDown(sqrtPX96, liquidity, amountOut, False)
        if zeroForOne
        else getNextSqrtPriceFromAmount0RoundingUp(sqrtPX96, liquidity, amountOut, False)
    )



from typing import List

from scipy.optimize import minimize_scalar

from ..baseclasses import BaseArbitrage
from ..erc20_token import Erc20Token
from ..uniswap.v2_dataclasses import UniswapV2PoolState
from ..uniswap.v2_liquidity_pool import LiquidityPool

# TODO: improve arbitrage calculation for repaying with same token, instead of borrow A -> repay B


class FlashBorrowToLpSwapNew(BaseArbitrage):
    def __init__(
        self,
        borrow_pool: LiquidityPool,
        borrow_token: Erc20Token,
        repay_token: Erc20Token,
        swap_pool_addresses: List[str] | None = None,
        swap_pools: List[LiquidityPool] | None = None,
        name: str = "",
        update_method: str = "polling",
    ):
        if swap_pools is None and swap_pool_addresses is None:
            raise TypeError("Provide a list of LiquidityPool objects or a list of pool addresses.")

        if not (swap_pools is not None) ^ (swap_pool_addresses is not None):
            # NOTE: ^ is the exclusive-or operator, which allows us to check that only one condition is True, but not both and not neither
            raise TypeError(
                "Provide a list of LiquidityPool objects or a list of pool addresses, but not both."
            )

        if update_method not in [
            "polling",
            "external",
        ]:
            raise ValueError("update_method must be 'polling' or 'external'")

        if update_method == "external" and swap_pool_addresses:
            raise ValueError(
                "swap pools passed by address must be updated with the 'polling' method"
            )

        if swap_pool_addresses is None:
            swap_pool_addresses = []

        if swap_pools is None:
            swap_pools = []

        self.borrow_pool = borrow_pool
        self.borrow_token = borrow_token
        self.repay_token = repay_token
        self._update_method = update_method

        # if the object was initialized with pool objects directly, use these directly
        self.swap_pools: List[LiquidityPool]
        if swap_pools:
            self.swap_pools = swap_pools

        # otherwise, create internal objects
        else:
            self.swap_pools = []
            for address in swap_pool_addresses:
                self.swap_pools.append(
                    LiquidityPool(
                        address=address,
                        update_method=update_method,
                    )
                )

        self.swap_pool_addresses = [pool.address for pool in self.swap_pools]
        self.swap_pool_tokens = [[pool.token0, pool.token1] for pool in self.swap_pools]

        self.all_pool_addresses = [pool.address for pool in self.swap_pools + [self.borrow_pool]]

        if name:
            self.name = name
        else:
            self.name = (
                self.borrow_pool.name
                + " -> "
                + " -> ".join([pool.name for pool in self.swap_pools])
            )

        if self.borrow_token.address not in [
            self.swap_pools[0].token0.address,
            self.swap_pools[0].token1.address,
        ]:
            raise ValueError("Borrowed token not found in the first swap pool")

        if self.repay_token.address not in [
            self.swap_pools[-1].token0.address,
            self.swap_pools[-1].token1.address,
        ]:
            raise ValueError("Repay token not found in the last swap pool")

        if self.swap_pools[0].token0 == borrow_token:
            forward_token_address = self.swap_pools[0].token1.address
        else:
            forward_token_address = self.swap_pools[0].token0.address

        token_in_address = forward_token_address
        for pool in self.swap_pools:
            if pool.token0 == token_in_address:
                forward_token_address = pool.token1.address
            elif pool.token1 == token_in_address:
                forward_token_address = pool.token0.address
            else:
                raise Exception("Swap pools are invalid, no swap route possible!")

        self.best = {
            "init": True,
            "strategy": "flash borrow swap",
            "borrow_amount": 0,
            "borrow_token": self.borrow_token,
            "borrow_pool": self.borrow_pool,
            "borrow_pool_amounts": [],
            "repay_amount": 0,
            "profit_amount": 0,
            "profit_token": self.repay_token,
            "swap_pools": self.swap_pools,
            "swap_pool_addresses": self.swap_pool_addresses,
            "swap_pool_amounts": [],
            "swap_pool_tokens": self.swap_pool_tokens,
        }

        self.best_future = {
            "strategy": "flash borrow swap",
            "borrow_amount": 0,
            "borrow_token": self.borrow_token,
            "borrow_pool": self.borrow_pool,
            "borrow_pool_amounts": [],
            "repay_amount": 0,
            "profit_amount": 0,
            "profit_token": self.repay_token,
            "swap_pools": self.swap_pools,
            "swap_pool_addresses": self.swap_pool_addresses,
            "swap_pool_amounts": [],
            "swap_pool_tokens": self.swap_pool_tokens,
        }

        self.update_reserves()

    def _build_multipool_amounts_out(
        self,
        token_in: Erc20Token,
        token_in_quantity: int,
    ) -> List[List[int]]:
        number_of_pools = len(self.swap_pools)

        pools_amounts_out = []

        for i in range(number_of_pools):
            # determine the output token for pool0
            if token_in.address == self.swap_pools[i].token0.address:
                token_out = self.swap_pools[i].token1
            elif token_in.address == self.swap_pools[i].token1.address:
                token_out = self.swap_pools[i].token0
            else:
                raise ValueError(f"Could not identify token_in {token_in}")

            # calculate the swap output through pool[i]
            token_out_quantity = self.swap_pools[i].calculate_tokens_out_from_tokens_in(
                token_in=token_in,
                token_in_quantity=token_in_quantity,
            )

            if token_in.address == self.swap_pools[i].token0.address:
                pools_amounts_out.append([0, token_out_quantity])
            elif token_in.address == self.swap_pools[i].token1.address:
                pools_amounts_out.append([token_out_quantity, 0])

            if i == number_of_pools - 1:
                break
            else:
                # otherwise, feed the results back into the loop
                token_in = token_out
                token_in_quantity = token_out_quantity

        return pools_amounts_out

    def _calculate_arbitrage(
        self,
        override_future: bool = False,
        override_future_borrow_pool_reserves_token0: int | None = None,
        override_future_borrow_pool_reserves_token1: int | None = None,
    ) -> None:
        if override_future:
            if (
                override_future_borrow_pool_reserves_token0 is None
                or override_future_borrow_pool_reserves_token1 is None
            ):
                raise ValueError("Must override reserves for token0 and token1")
            reserves_token0 = override_future_borrow_pool_reserves_token0
            reserves_token1 = override_future_borrow_pool_reserves_token1
        else:
            if (
                override_future_borrow_pool_reserves_token0 is not None
                or override_future_borrow_pool_reserves_token1 is not None
            ):
                raise ValueError(
                    "Do not provide override reserves without setting override_future = True"
                )
            reserves_token0 = self.borrow_pool.reserves_token0
            reserves_token1 = self.borrow_pool.reserves_token1

        # set up the boundaries for the Brent optimizer based on which token is being borrowed
        if self.borrow_token == self.borrow_pool.token0:
            bounds = (
                1,
                float(reserves_token0),
            )
            bracket = (
                0.001 * reserves_token0,
                0.01 * reserves_token0,
            )
        elif self.borrow_token == self.borrow_pool.token1:
            bounds = (
                1,
                float(reserves_token1),
            )
            bracket = (
                0.001 * reserves_token1,
                0.01 * reserves_token1,
            )
        else:
            raise ValueError(f"Could not identify borrow token {self.borrow_token}")

        # TODO: extend calculate_multipool_tokens_out_from_tokens_in() to support overriding token reserves for an arbitrary pool,
        # currently only supports overriding the borrow pool reserves
        opt = minimize_scalar(
            lambda x: -float(
                self.calculate_multipool_tokens_out_from_tokens_in(
                    token_in=self.borrow_token,
                    token_in_quantity=x,
                )
                - self.borrow_pool.calculate_tokens_in_from_tokens_out(
                    token_out=(
                        self.borrow_pool.token0
                        if self.repay_token is self.borrow_pool.token1
                        else self.borrow_pool.token1
                    ),
                    token_out_quantity=x,
                    override_state=UniswapV2PoolState(
                        pool=self.borrow_pool,
                        reserves_token0=reserves_token0,
                        reserves_token1=reserves_token1,
                    ),
                )
            ),
            method="bounded",
            bounds=bounds,
            bracket=bracket,
            options={"xatol": 1.0},
        )

        best_borrow = int(opt.x)

        if self.borrow_token == self.borrow_pool.token0:
            borrow_amounts = [best_borrow, 0]
        elif self.borrow_token == self.borrow_pool.token1:
            borrow_amounts = [0, best_borrow]
        else:
            raise ValueError(f"Could not identify borrow token {self.borrow_token}")

        best_repay = self.borrow_pool.calculate_tokens_in_from_tokens_out(
            token_out=(
                self.borrow_pool.token0
                if self.repay_token is self.borrow_pool.token1
                else self.borrow_pool.token1
            ),
            token_out_quantity=best_borrow,
            override_state=UniswapV2PoolState(
                pool=self.borrow_pool,
                reserves_token0=reserves_token0,
                reserves_token1=reserves_token1,
            ),
        )
        best_profit = -int(opt.fun)

        if override_future:
            if best_borrow > 0 and best_profit > 0:
                self.best_future.update(
                    {
                        "borrow_amount": best_borrow,
                        "borrow_pool_amounts": borrow_amounts,
                        "repay_amount": best_repay,
                        "profit_amount": best_profit,
                        "swap_pool_amounts": self._build_multipool_amounts_out(
                            token_in=self.borrow_token,
                            token_in_quantity=best_borrow,
                        ),
                    }
                )
            else:
                self.best_future.update(
                    {
                        "borrow_amount": 0,
                        "borrow_pool_amounts": [],
                        "repay_amount": 0,
                        "profit_amount": 0,
                        "swap_pool_amounts": [],
                    }
                )
        else:
            # only save opportunities with rational, positive values
            if best_borrow > 0 and best_profit > 0:
                self.best.update(
                    {
                        "borrow_amount": best_borrow,
                        "borrow_pool_amounts": borrow_amounts,
                        "repay_amount": best_repay,
                        "profit_amount": best_profit,
                        "swap_pool_amounts": self._build_multipool_amounts_out(
                            token_in=self.borrow_token,
                            token_in_quantity=best_borrow,
                        ),
                    }
                )
            else:
                self.best.update(
                    {
                        "borrow_amount": 0,
                        "borrow_pool_amounts": [],
                        "repay_amount": 0,
                        "profit_amount": 0,
                        "swap_pool_amounts": [],
                    }
                )

    def __str__(self) -> str:
        return self.name

    def calculate_multipool_tokens_out_from_tokens_in(
        self, token_in: Erc20Token, token_in_quantity: int
    ) -> int:
        """
        Calculates the expected token OUTPUT from the last pool for a given token INPUT to the first pool at current pool reserves.
        Uses the self.token0 and self.token1 pointers to determine which token is being swapped in
        and uses the appropriate formula
        """

        number_of_pools = len(self.swap_pools)

        for i in range(number_of_pools):
            # determine the output token for pool0
            if token_in.address == self.swap_pools[i].token0.address:
                token_out = self.swap_pools[i].token1
            elif token_in.address == self.swap_pools[i].token1.address:
                token_out = self.swap_pools[i].token0
            else:
                raise ValueError(f"Could not identify token_in {token_in}")

            # calculate the swap output through pool[i]
            token_out_quantity = self.swap_pools[i].calculate_tokens_out_from_tokens_in(
                token_in=token_in,
                token_in_quantity=token_in_quantity,
            )

            if i == number_of_pools - 1:
                break
            else:
                # otherwise, use the output as input on the next loop
                token_in = token_out
                token_in_quantity = token_out_quantity

        return token_out_quantity

    def update_reserves(
        self,
        silent: bool = False,
        print_reserves: bool = True,
        print_ratios: bool = True,
        override_future: bool = False,
        override_future_borrow_pool_reserves_token0: int | None = None,
        override_future_borrow_pool_reserves_token1: int | None = None,
    ) -> bool:
        """
        Checks each liquidity pool for updates by passing a call to .update_reserves(), which returns False if there are no updates.
        Will calculate arbitrage amounts only after checking all pools and finding a reason to update, or on startup (via the 'init' dictionary key)
        """
        recalculate = False

        if self._update_method != "external":
            # calculate initial arbitrage after the object is instantiated, otherwise proceed with normal checks
            if self.best["init"] is True:
                self.best["init"] = False
                recalculate = True

            # flag for recalculation if the borrowing pool has been updated
            if self.borrow_pool.update_reserves(
                silent=silent,
                print_reserves=print_reserves,
                print_ratios=print_ratios,
            ):
                recalculate = True

            # flag for recalculation if any of the pools along the swap path have been updated
            for pool in self.swap_pools:
                if pool.update_reserves(
                    silent=silent,
                    print_reserves=print_reserves,
                    print_ratios=print_ratios,
                ):
                    recalculate = True

        if override_future:
            recalculate = True
            if (
                override_future_borrow_pool_reserves_token0 is None
                or override_future_borrow_pool_reserves_token1 is None
            ):
                raise ValueError("Must override reserves for token0 and token1")
        else:
            if (
                override_future_borrow_pool_reserves_token0 is not None
                or override_future_borrow_pool_reserves_token1 is not None
            ):
                raise ValueError(
                    "Do not provide override reserves without setting override_future = True"
                )

        if self.borrow_pool.new_reserves:
            recalculate = True
        for pool in self.swap_pools:
            if pool.new_reserves:
                recalculate = True
                break

        if recalculate:
            self._calculate_arbitrage(
                override_future=override_future,
                override_future_borrow_pool_reserves_token0=override_future_borrow_pool_reserves_token0,
                override_future_borrow_pool_reserves_token1=override_future_borrow_pool_reserves_token1,
            )
            return True
        else:
            return False



from typing import Dict
from eth_utils.address import to_checksum_address
from eth_typing import ChecksumAddress

from ..baseclasses import BaseToken


_all_tokens: Dict[
    int,
    Dict[ChecksumAddress, BaseToken],
] = {}


class AllTokens:
    def __init__(self, chain_id: int) -> None:
        try:
            _all_tokens[chain_id]
        except KeyError:
            _all_tokens[chain_id] = {}
        finally:
            self.tokens = _all_tokens[chain_id]

    def __delitem__(self, token_address: str) -> None:
        del self.tokens[to_checksum_address(token_address)]

    def __getitem__(self, token_address: str) -> BaseToken:
        return self.tokens[to_checksum_address(token_address)]

    def __setitem__(self, token_address: str, token_helper: BaseToken) -> None:
        self.tokens[to_checksum_address(token_address)] = token_helper

    def __len__(self) -> int:
        return len(self.tokens)

    def get(self, token_address: str) -> BaseToken | None:
        return self.tokens.get(to_checksum_address(token_address))



from typing import Any, Dict, List, Tuple
from warnings import warn

import eth_abi.abi
import ujson
from eth_typing import AnyAddress, ChecksumAddress
from eth_utils.address import to_checksum_address
from web3 import Web3
from web3.contract.contract import Contract
from web3.exceptions import BadFunctionCallOutput, ContractLogicError
from web3.types import BlockIdentifier

from . import config
from .baseclasses import BaseToken
from .chainlink import ChainlinkPriceContract
from .functions import get_number_for_block_identifier
from .logging import logger
from .registry.all_tokens import AllTokens

# Taken from OpenZeppelin's ERC-20 implementation
# ref: https://www.npmjs.com/package/@openzeppelin/contracts?activeTab=code
ERC20_ABI_MINIMAL = ujson.loads(
    '[{"inputs": [{"internalType": "string", "name": "name_", "type": "string"}, {"internalType": "string", "name": "symbol_", "type": "string"}], "stateMutability": "nonpayable", "type": "constructor"}, {"anonymous": false, "inputs": [{"indexed": true, "internalType": "address", "name": "owner", "type": "address"}, {"indexed": true, "internalType": "address", "name": "spender", "type": "address"}, {"indexed": false, "internalType": "uint256", "name": "value", "type": "uint256"}], "name": "Approval", "type": "event"}, {"anonymous": false, "inputs": [{"indexed": true, "internalType": "address", "name": "from", "type": "address"}, {"indexed": true, "internalType": "address", "name": "to", "type": "address"}, {"indexed": false, "internalType": "uint256", "name": "value", "type": "uint256"}], "name": "Transfer", "type": "event"}, {"inputs": [{"internalType": "address", "name": "owner", "type": "address"}, {"internalType": "address", "name": "spender", "type": "address"}], "name": "allowance", "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}], "stateMutability": "view", "type": "function"}, {"inputs": [{"internalType": "address", "name": "spender", "type": "address"}, {"internalType": "uint256", "name": "amount", "type": "uint256"}], "name": "approve", "outputs": [{"internalType": "bool", "name": "", "type": "bool"}], "stateMutability": "nonpayable", "type": "function"}, {"inputs": [{"internalType": "address", "name": "account", "type": "address"}], "name": "balanceOf", "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}], "stateMutability": "view", "type": "function"}, {"inputs": [], "name": "decimals", "outputs": [{"internalType": "uint8", "name": "", "type": "uint8"}], "stateMutability": "view", "type": "function"}, {"inputs": [{"internalType": "address", "name": "spender", "type": "address"}, {"internalType": "uint256", "name": "subtractedValue", "type": "uint256"}], "name": "decreaseAllowance", "outputs": [{"internalType": "bool", "name": "", "type": "bool"}], "stateMutability": "nonpayable", "type": "function"}, {"inputs": [{"internalType": "address", "name": "spender", "type": "address"}, {"internalType": "uint256", "name": "addedValue", "type": "uint256"}], "name": "increaseAllowance", "outputs": [{"internalType": "bool", "name": "", "type": "bool"}], "stateMutability": "nonpayable", "type": "function"}, {"inputs": [], "name": "name", "outputs": [{"internalType": "string", "name": "", "type": "string"}], "stateMutability": "view", "type": "function"}, {"inputs": [], "name": "symbol", "outputs": [{"internalType": "string", "name": "", "type": "string"}], "stateMutability": "view", "type": "function"}, {"inputs": [], "name": "totalSupply", "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}], "stateMutability": "view", "type": "function"}, {"inputs": [{"internalType": "address", "name": "to", "type": "address"}, {"internalType": "uint256", "name": "amount", "type": "uint256"}], "name": "transfer", "outputs": [{"internalType": "bool", "name": "", "type": "bool"}], "stateMutability": "nonpayable", "type": "function"}, {"inputs": [{"internalType": "address", "name": "from", "type": "address"}, {"internalType": "address", "name": "to", "type": "address"}, {"internalType": "uint256", "name": "amount", "type": "uint256"}], "name": "transferFrom", "outputs": [{"internalType": "bool", "name": "", "type": "bool"}], "stateMutability": "nonpayable", "type": "function"}]'
)


class Erc20Token(BaseToken):
    """
    An ERC-20 token contract.

    If an ABI is specified, it will be used. Otherwise, a minimal ERC-20 ABI
    will be used that provides access to the interface defined by EIP-20, plus
    some commonly used extensions:

    - Functions:
        # EIP-20 BASELINE
        - totalSupply()
        - balanceOf(account)
        - transfer(to, amount)
        - allowance(owner, spender)
        - approve(spender, amount)
        - transferFrom(from, to, amount)
        # METADATA EXTENSIONS
        - name()
        - symbol()
        - decimals()
        # ALLOWANCE EXTENSIONS
        - increaseAllowance(spender, addedValue)
        - decreaseAllowance(spender, subtractedValue)
    - Events:
        - Transfer(from, to, value)
        - Approval(owner, spender, value)
    """

    def __init__(
        self,
        address: str,
        abi: List[Any] | None = None,
        oracle_address: str | None = None,
        silent: bool = False,
        unload_brownie_contract_after_init: bool = False,  # deprecated
        min_abi: bool = False,  # deprecated
        user: Any | None = None,  # deprecated
    ) -> None:
        self.address: ChecksumAddress = to_checksum_address(address)
        self.abi = abi if abi is not None else ERC20_ABI_MINIMAL

        _w3 = config.get_web3()
        _w3_contract = self._w3_contract

        if user:  # pragma: no cover
            warn(
                "Instantiating with a single user is deprecated. You may use "
                "the get_balance() method to retrieve token balances for a "
                "particular address."
            )

        if min_abi:  # pragma: no cover
            warn(
                "Using a minimal ABI is now the default behavior. Remove "
                "min_abi constructor argument to stop seeing this message."
            )

        if unload_brownie_contract_after_init:  # pragma: no cover
            warn(
                "unload_brownie_contract_after_init is deprecated. Remove "
                "constructor argument to stop seeing this message."
            )

        try:
            self.name: str
            self.name = _w3_contract.functions.name().call()
        except (ContractLogicError, OverflowError, BadFunctionCallOutput):
            # Workaround for non-ERC20 compliant tokens
            for func in ("name", "NAME"):
                try:
                    self.name = (
                        _w3.eth.call(
                            {
                                "to": self.address,
                                "data": Web3.keccak(text=f"{func}()"),
                            }
                        )
                    ).decode("utf-8", errors="ignore")
                except Exception:
                    continue
                else:
                    break
        except Exception as e:
            print(f"(token.name @ {self.address}) {type(e)}: {e}")
            raise

        try:
            self.name
        except AttributeError:  # pragma: no cover
            if not _w3.eth.get_code(self.address):
                raise ValueError("No contract deployed at this address")
            self.name = f"Unknown @ {self.address}"
            warn(
                f"Token contract at {self.address} does not implement a 'name' function. Setting to '{self.name}'"
            )
        finally:
            self.name = self.name.strip("\x00")

        try:
            self.symbol: str
            self.symbol = _w3_contract.functions.symbol().call()
        except (ContractLogicError, OverflowError, BadFunctionCallOutput):
            for func in ("symbol", "SYMBOL"):
                # Workaround for non-ERC20 compliant tokens
                try:
                    self.symbol = (
                        _w3.eth.call(
                            {
                                "to": self.address,
                                "data": Web3.keccak(text=f"{func}()"),
                            }
                        )
                    ).decode("utf-8", errors="ignore")
                except Exception:
                    continue
                else:
                    break
        except Exception as e:
            print(f"(token.symbol @ {self.address}) {type(e)}: {e}")
            raise

        try:
            self.symbol
        except AttributeError:
            if not _w3.eth.get_code(self.address):  # pragma: no cover
                raise ValueError("No contract deployed at this address")
            self.symbol = "UNKN"
            warn(
                f"Token contract at {self.address} does not implement a 'symbol' function. Setting to {self.symbol}"
            )

        try:
            self.decimals: int
            self.decimals = _w3_contract.functions.decimals().call()
        except (ContractLogicError, OverflowError, BadFunctionCallOutput):
            for func in ("decimals", "DECIMALS"):
                try:
                    # Workaround for non-ERC20 compliant tokens
                    self.decimals = int.from_bytes(
                        bytes=_w3.eth.call(
                            {
                                "to": self.address,
                                "data": Web3.keccak(text=f"{func}()"),
                            }
                        ),
                        byteorder="big",
                    )
                except Exception:
                    continue
                else:
                    break
        except Exception as e:
            print(f"(token.decimals @ {self.address}) {type(e)}: {e}")
            raise

        try:
            self.decimals
        except Exception:
            if not _w3.eth.get_code(self.address):  # pragma: no cover
                raise ValueError("No contract deployed at this address")
            self.decimals = 0
            warn(
                f"Token contract at {self.address} does not implement a 'decimals' function. Setting to 0."
            )

        self.price: float | None = None
        if oracle_address:
            self._price_oracle = ChainlinkPriceContract(address=oracle_address)
            self.price = self._price_oracle.price
        else:
            self.price = None

        AllTokens(chain_id=_w3.eth.chain_id)[self.address] = self

        self._cached_approval: Dict[Tuple[int, ChecksumAddress, ChecksumAddress], int] = {}
        self._cached_balance: Dict[Tuple[int, ChecksumAddress], int] = {}
        self._cached_total_supply: Dict[int, int] = {}

        if not silent:
            logger.info(f"• {self.symbol} ({self.name})")

    def __repr__(self) -> str:  # pragma: no cover
        return f"Erc20Token(address={self.address}, symbol='{self.symbol}', name='{self.name}', decimals={self.decimals})"

    @property
    def _w3_contract(self) -> Contract:
        return config.get_web3().eth.contract(
            address=self.address,
            abi=self.abi,
        )

    def _get_approval_cachable(
        self,
        owner: ChecksumAddress,
        spender: ChecksumAddress,
        block_number: int,
    ) -> int:
        try:
            return self._cached_approval[block_number, owner, spender]
        except KeyError:
            pass

        approval: int
        approval, *_ = eth_abi.abi.decode(
            types=["uint256"],
            data=config.get_web3().eth.call(
                transaction={
                    "to": self.address,
                    "data": Web3.keccak(text="allowance(address,address)")[:4]
                    + eth_abi.abi.encode(types=["address", "address"], args=[owner, spender]),
                },
                block_identifier=block_number,
            ),
        )
        self._cached_approval[block_number, owner, spender] = approval
        return approval

    def get_approval(
        self,
        owner: AnyAddress,
        spender: AnyAddress,
        block_identifier: BlockIdentifier | None = None,
    ) -> int:
        """
        Retrieve the amount that can be spent by `spender` on behalf of `owner`.
        """

        return self._get_approval_cachable(
            to_checksum_address(owner),
            to_checksum_address(spender),
            block_number=get_number_for_block_identifier(block_identifier),
        )

    def _get_balance_cachable(
        self,
        address: ChecksumAddress,
        block_number: int,
    ) -> int:
        try:
            return self._cached_balance[block_number, address]
        except KeyError:
            pass

        balance: int
        balance, *_ = eth_abi.abi.decode(
            types=["uint256"],
            data=config.get_web3().eth.call(
                transaction={
                    "to": self.address,
                    "data": Web3.keccak(text="balanceOf(address)")[:4]
                    + eth_abi.abi.encode(types=["address"], args=[address]),
                },
                block_identifier=block_number,
            ),
        )
        self._cached_balance[block_number, address] = balance
        return balance

    def get_balance(
        self,
        address: AnyAddress,
        block_identifier: BlockIdentifier | None = None,
    ) -> int:
        """
        Retrieve the ERC-20 balance for the given address.
        """
        return self._get_balance_cachable(
            address=to_checksum_address(address),
            block_number=get_number_for_block_identifier(block_identifier),
        )

    def _get_total_supply_cachable(self, block_number: int) -> int:
        try:
            return self._cached_total_supply[block_number]
        except KeyError:
            pass

        total_supply: int
        total_supply, *_ = eth_abi.abi.decode(
            types=["uint256"],
            data=config.get_web3().eth.call(
                transaction={"to": self.address, "data": Web3.keccak(text="totalSupply()")[:4]},
                block_identifier=block_number,
            ),
        )
        self._cached_total_supply[block_number] = total_supply
        return total_supply

    def get_total_supply(self, block_identifier: BlockIdentifier | None = None) -> int:
        """
        Retrieve the total supply for this token.
        """

        if block_identifier is None:
            block_identifier = config.get_web3().eth.get_block_number()

        return self._get_total_supply_cachable(
            block_number=get_number_for_block_identifier(block_identifier)
        )

    def update_price(self) -> None:
        self._price_oracle.update_price()
        self.price = self._price_oracle.price


class EEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEE(Erc20Token):
    """
    An adapter for pools using the 'all Es' placeholder address to represent native Ether.
    """

    address = to_checksum_address("0xEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEE")
    symbol = "ETH"
    name = "Ether Placeholder"
    decimals = 18

    def __init__(self) -> None:
        self._cached_balance: Dict[Tuple[int, ChecksumAddress], int] = {}
        AllTokens(chain_id=config.get_web3().eth.chain_id)[self.address] = self

    def _get_balance_cachable(
        self,
        address: ChecksumAddress,
        block_number: int,
    ) -> int:
        try:
            return self._cached_balance[block_number, address]
        except KeyError:
            pass

        balance = config.get_web3().eth.get_balance(
            address,
            block_identifier=block_number,
        )
        self._cached_balance[block_number, address] = balance
        return balance

    def get_balance(
        self,
        address: AnyAddress,
        block_identifier: BlockIdentifier | None = None,
    ) -> int:
        return self._get_balance_cachable(
            address=to_checksum_address(address),
            block_number=get_number_for_block_identifier(block_identifier),
        )



import degenbot
from degenbot.exceptions import DegenbotError
import web3
import pytest


def test_disconnected_web3():
    w3 = web3.Web3(web3.HTTPProvider("https://google.com"))
    with pytest.raises(DegenbotError, match="Web3 object is not connected."):
        degenbot.set_web3(w3)


def test_unset_web3():
    WETH_ADDRESS = "0xC02aaA39b223FE8D0A0e5C4F27eAD9083C756Cc2"

    del degenbot.config._web3

    with pytest.raises(DegenbotError, match="A Web3 instance has not been provided."):
        degenbot.Erc20Token(WETH_ADDRESS)



from degenbot.functions import next_base_fee, get_number_for_block_identifier
from degenbot.constants import MAX_UINT256
import pytest
import degenbot.config
from eth_typing import (
    BlockNumber,
    Hash32,
    HexStr,
)
from hexbytes import HexBytes


def test_convert_identifier(fork_mainnet_archive):
    """
    Check that all inputs for type BlockIdentifier can be converted to an integer
    """

    degenbot.config.set_web3(fork_mainnet_archive.w3)
    for block_identifier in [
        "latest",  # BlockParams
        "earliest",  # BlockParams
        "pending",  # BlockParams
        "safe",  # BlockParams
        "finalized",  # BlockParams
        BlockNumber(1),  # BlockNumber
        Hash32(int(1).to_bytes(length=32, byteorder="big")),  # Hash32
        HexStr("0x" + int(128).to_bytes(32, byteorder="big").hex()),  # HexStr
        HexBytes(1),  # HexBytes
        1,  # int
    ]:
        assert isinstance(get_number_for_block_identifier(block_identifier), int)

    for invalid_block_number in [-1, MAX_UINT256 + 1]:
        with pytest.raises(ValueError):
            get_number_for_block_identifier(invalid_block_number)

    for invalid_tag in ["Latest", "latest ", "next", "previous"]:
        with pytest.raises(ValueError):
            get_number_for_block_identifier(invalid_tag)


def test_fee_calcs():
    BASE_FEE = 100 * 10**9

    # EIP-1559 target is 50% full blocks, so a 50% full block should return the same base fee
    assert (
        next_base_fee(
            parent_base_fee=BASE_FEE,
            parent_gas_used=15_000_000,
            parent_gas_limit=30_000_000,
        )
        == BASE_FEE
    )

    # Fee should be higher
    assert (
        next_base_fee(
            parent_base_fee=BASE_FEE,
            parent_gas_used=20_000_000,
            parent_gas_limit=30_000_000,
        )
        == 104166666666
    )

    # Fee should be lower
    assert (
        next_base_fee(
            parent_base_fee=BASE_FEE,
            parent_gas_used=10_000_000,
            parent_gas_limit=30_000_000,
        )
        == 95833333334
    )

    MIN_BASE_FEE = 95 * 10**9

    # Enforce minimum fee
    assert (
        next_base_fee(
            parent_base_fee=BASE_FEE,
            parent_gas_used=0,
            parent_gas_limit=30_000_000,
            min_base_fee=MIN_BASE_FEE,
        )
        == MIN_BASE_FEE
    )



from threading import Lock
from typing import Any, Dict

from eth_typing import ChecksumAddress
from eth_utils.address import to_checksum_address

from .. import config
from ..baseclasses import BaseManager
from ..erc20_token import EEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEE, Erc20Token
from ..exceptions import ManagerError


class Erc20TokenHelperManager(BaseManager):
    """
    A class that generates and tracks Erc20Token helpers

    The state dictionary is held using the "Borg" singleton pattern, which
    ensures that all instances of the class have access to the same state data
    """

    _state: Dict[int, Dict[str, Any]] = {}

    def __init__(self, chain_id: int | None = None) -> None:
        chain_id = chain_id if chain_id is not None else config.get_web3().eth.chain_id

        # the internal state data for this object is held in the
        # class-level _state dictionary, keyed by the chain ID
        if self._state.get(chain_id):
            self.__dict__ = self._state[chain_id]
        else:
            self._state[chain_id] = {}
            self.__dict__ = self._state[chain_id]

            # initialize internal attributes
            self._erc20tokens: Dict[ChecksumAddress, Erc20Token] = {}
            self._lock = Lock()

    def get_erc20token(
        self,
        address: str,
        # accept any number of keyword arguments, which are
        # passed directly to Erc20Token without validation
        **kwargs: Any,
    ) -> Erc20Token:
        """
        Get the token object from its address
        """

        address = to_checksum_address(address)

        if token_helper := self._erc20tokens.get(address):
            return token_helper

        try:
            if address == EEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEE.address:
                token_helper = EEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEE()
            else:
                token_helper = Erc20Token(address=address, **kwargs)
        except Exception:
            raise ManagerError(f"Could not create Erc20Token helper: {address=}")

        with self._lock:
            self._erc20tokens[address] = token_helper

        return token_helper



from threading import Lock
from typing import TYPE_CHECKING, Any, Dict, List, Set, Tuple

from eth_typing import ChecksumAddress
from eth_utils.address import to_checksum_address
from web3.contract.contract import Contract

from .. import config
from ..baseclasses import BaseManager
from ..constants import ZERO_ADDRESS
from ..dex.uniswap import FACTORY_ADDRESSES, TICKLENS_ADDRESSES
from ..erc20_token import Erc20Token
from ..exceptions import ManagerError, PoolNotAssociated
from ..logging import logger
from ..manager.token_manager import Erc20TokenHelperManager
from ..registry.all_pools import AllPools
from .abi import UNISWAP_V2_FACTORY_ABI
from .v2_liquidity_pool import LiquidityPool
from .v3_functions import generate_v3_pool_address
from .v3_liquidity_pool import V3LiquidityPool
from .v3_snapshot import UniswapV3LiquiditySnapshot
from .v3_tick_lens import TickLens


class UniswapLiquidityPoolManager(BaseManager):
    """
    Single-concern base class to allow derived classes to share state
    """

    _state: Dict[int, Dict[str, Any]] = dict()

    def __init__(
        self,
        factory_address: str,
        chain_id: int,
    ):
        """
        Initialize the specific state dictionary for the given chain id and
        factory address
        """

        # the internal state data for all child objects is held in a nested
        # class-level dictionary, keyed by chain ID and factory address
        try:
            self._state[chain_id]
        except KeyError:
            self._state[chain_id] = {}
            self._state[chain_id]["erc20token_manager"] = Erc20TokenHelperManager(chain_id)

        try:
            self._state[chain_id][factory_address]
        except KeyError:
            self._state[chain_id][factory_address] = {}

    @classmethod
    def add_chain(cls, chain_id: int) -> None:
        """
        Add a new chain ID.
        """
        if not FACTORY_ADDRESSES.get(chain_id):
            FACTORY_ADDRESSES[chain_id] = {}

    @classmethod
    def add_factory(cls, chain_id: int, factory_address: str) -> None:
        """
        Add a new factory address at a given chain ID.
        """
        cls.add_chain(chain_id=chain_id)

        factory_address = to_checksum_address(factory_address)

        if not FACTORY_ADDRESSES[chain_id].get(factory_address):
            FACTORY_ADDRESSES[chain_id][factory_address] = {}

    @classmethod
    def add_pool_init_hash(cls, chain_id: int, factory_address: str, pool_init_hash: str) -> None:
        """
        Add a pool_init_hash for a factory at a given chain ID.
        """

        factory_address = to_checksum_address(factory_address)

        cls.add_factory(chain_id=chain_id, factory_address=factory_address)

        if not FACTORY_ADDRESSES[chain_id][factory_address].get("init_hash"):
            FACTORY_ADDRESSES[chain_id][factory_address]["init_hash"] = pool_init_hash


class UniswapV2LiquidityPoolManager(UniswapLiquidityPoolManager):
    """
    A class that generates and tracks Uniswap V2 liquidity pool helpers

    The state dictionary is held using the "Borg" singleton pattern, which
    ensures that all instances of the class have access to the same state data
    """

    def __init__(
        self,
        factory_address: ChecksumAddress | str,
        chain_id: int | None = None,
    ):
        chain_id = chain_id if chain_id is not None else config.get_web3().eth.chain_id

        factory_address = to_checksum_address(factory_address)

        if factory_address not in FACTORY_ADDRESSES[chain_id]:
            raise ManagerError(
                f"Pool manager could not be initialized from unknown factory address {factory_address}. Add the factory address and pool init hash with `add_factory`, followed by `add_pool_init_hash`"
            )

        super().__init__(
            factory_address=factory_address,
            chain_id=chain_id,
        )

        self.__dict__ = self._state[chain_id][factory_address]

        if self.__dict__ == {}:
            try:
                self.chain_id = chain_id
                self._factory_address = factory_address
                self._lock = Lock()
                self._tracked_pools: Dict[ChecksumAddress, LiquidityPool] = dict()
                self._token_manager: Erc20TokenHelperManager = self._state[chain_id][
                    "erc20token_manager"
                ]
                self._factory_init_hash = FACTORY_ADDRESSES[chain_id][self._factory_address][
                    "init_hash"
                ]
                self._untracked_pools: Set[ChecksumAddress] = set()
            except Exception as e:
                self._state[chain_id][factory_address] = {}
                raise ManagerError(f"Could not initialize state for {factory_address}") from e

    def __delitem__(self, pool: LiquidityPool | ChecksumAddress | str) -> None:
        pool_address: ChecksumAddress

        if isinstance(pool, LiquidityPool):
            pool_address = pool.address
        else:
            pool_address = to_checksum_address(pool)

        try:
            del self._tracked_pools[pool_address]
        except KeyError:
            pass

        self._untracked_pools.discard(pool_address)
        assert pool_address not in self._untracked_pools

    def __repr__(self) -> str:  # pragma: no cover
        return f"UniswapV2LiquidityPoolManager(factory={self._factory_address})"

    @property
    def _w3_contract(self) -> Contract:
        return config.get_web3().eth.contract(
            address=self._factory_address,
            abi=UNISWAP_V2_FACTORY_ABI,
        )

    def _add_pool(self, pool_helper: LiquidityPool) -> None:
        with self._lock:
            self._tracked_pools[pool_helper.address] = pool_helper
        assert pool_helper.address in self._tracked_pools

    def get_pool(
        self,
        pool_address: str | None = None,
        token_addresses: Tuple[str, str] | None = None,
        silent: bool = False,
        update_method: str = "polling",
        state_block: int | None = None,
    ) -> LiquidityPool:
        """
        Get the pool object from its address, or a tuple of token addresses
        """

        if token_addresses is not None:
            if len(token_addresses) != 2:
                raise ValueError("Provide exactly two token addresses")

            checksummed_token_addresses = tuple(
                [to_checksum_address(token_address) for token_address in token_addresses]
            )

            try:
                for token_address in checksummed_token_addresses:
                    self._token_manager.get_erc20token(
                        address=token_address,
                        silent=silent,
                    )
            except Exception:
                raise ManagerError("Could not get both Erc20Token helpers")

            if pool_address == ZERO_ADDRESS:
                raise ManagerError("No V2 LP available")

            pool_address = to_checksum_address(
                self._w3_contract.functions.getPair(*checksummed_token_addresses).call()
            )

        if TYPE_CHECKING:
            assert pool_address is not None
        # Address is now known, check if the pool is already being tracked
        pool_address = to_checksum_address(pool_address)

        if pool_address in self._untracked_pools:
            raise PoolNotAssociated(
                f"Pool address {pool_address} not associated with factory {self._factory_address}"
            )

        try:
            return self._tracked_pools[pool_address]
        except KeyError:
            pass

        # Check if the AllPools collection already has this pool
        pool_helper = AllPools(self.chain_id).get(pool_address)
        if pool_helper:
            if TYPE_CHECKING:
                assert isinstance(pool_helper, LiquidityPool)
            if pool_helper.factory == self._factory_address:
                self._add_pool(pool_helper)
                return pool_helper
            else:
                self._untracked_pools.add(pool_address)
                raise PoolNotAssociated(f"Pool {pool_address} is not associated with this DEX")

        try:
            pool_helper = LiquidityPool(
                address=pool_address,
                silent=silent,
                state_block=state_block,
                factory_address=self._factory_address,
                factory_init_hash=self._factory_init_hash,
                update_method=update_method,
            )
        except Exception as e:
            self._untracked_pools.add(
                pool_address
            )  # <--- this may be the cause of the spurious PoolNotAssociated exception, if the pool helper creation fails for some reason
            # logger.error(f"Adding {pool_address} to untracked pools. Reason: {e}")
            raise ManagerError(f"Could not build V2 pool {pool_address}: {e}")
        else:
            self._add_pool(pool_helper)
            return pool_helper


class UniswapV3LiquidityPoolManager(UniswapLiquidityPoolManager):
    """
    A class that generates and tracks Uniswap V3 liquidity pool helpers

    The state dictionary is held using the "Borg" singleton pattern, which
    ensures that all instances of the class have access to the same state data
    """

    def __init__(
        self,
        factory_address: ChecksumAddress | str,
        chain_id: int | None = None,
        snapshot: UniswapV3LiquiditySnapshot | None = None,
    ):
        chain_id = chain_id if chain_id is not None else config.get_web3().eth.chain_id

        factory_address = to_checksum_address(factory_address)

        if factory_address not in FACTORY_ADDRESSES[chain_id]:
            raise ManagerError(
                f"Pool manager could not be initialized from unknown factory address {factory_address}. Add the factory address and pool init hash with `add_factory`, followed by `add_pool_init_hash`"
            )

        super().__init__(
            factory_address=factory_address,
            chain_id=chain_id,
        )

        self.__dict__ = self._state[chain_id][factory_address]

        if self.__dict__ == {}:
            try:
                self.chain_id = chain_id
                self._factory_address = to_checksum_address(factory_address)
                self._lens = TickLens(address=TICKLENS_ADDRESSES[chain_id][factory_address])
                self._lock = Lock()
                self._tracked_pools: Dict[ChecksumAddress, V3LiquidityPool] = {}
                self._token_manager: Erc20TokenHelperManager = self._state[chain_id][
                    "erc20token_manager"
                ]
                self._factory_init_hash = FACTORY_ADDRESSES[chain_id][self._factory_address][
                    "init_hash"
                ]
                self._snapshot = snapshot
                self._untracked_pools: Set[ChecksumAddress] = set()
            except Exception as e:
                self._state[chain_id][factory_address] = {}
                raise ManagerError(f"Could not initialize state for {factory_address}") from e

    def __delitem__(self, pool: V3LiquidityPool | ChecksumAddress | str) -> None:
        pool_address: ChecksumAddress

        if isinstance(pool, V3LiquidityPool):
            pool_address = pool.address
        else:
            pool_address = to_checksum_address(pool)

        try:
            del self._tracked_pools[pool_address]
        except KeyError:
            pass

        self._untracked_pools.discard(pool_address)

    def __repr__(self) -> str:  # pragma: no cover
        return f"UniswapV3LiquidityPoolManager(factory={self._factory_address})"

    def _add_pool(self, pool_helper: V3LiquidityPool) -> None:
        with self._lock:
            self._tracked_pools[pool_helper.address] = pool_helper

    def get_pool(
        self,
        pool_address: ChecksumAddress | str | None = None,
        token_addresses: Tuple[
            ChecksumAddress | str,
            ChecksumAddress | str,
        ]
        | None = None,
        pool_fee: int | None = None,
        silent: bool = False,
        # keyword arguments passed to the `V3LiquidityPool` constructor
        v3liquiditypool_kwargs: Dict[str, Any] | None = None,
        state_block: int | None = None,
    ) -> V3LiquidityPool:
        """
        Get a `V3LiquidityPool` from its address, or a tuple of token addresses
        and fee in bips (e.g. 100, 500, 3000, 10000)
        """

        def apply_liquidity_updates(pool: V3LiquidityPool) -> None:
            logger.debug(f"Applying liquidity updates to {pool}")
            if not self._snapshot:
                return
            for external_update in self._snapshot.get_pool_updates(pool.address):
                pool.external_update(update=external_update, force=True)

        def find_or_build(
            pool_address: ChecksumAddress,
            state_block: int | None = None,
        ) -> V3LiquidityPool:
            if TYPE_CHECKING:
                assert isinstance(v3liquiditypool_kwargs, dict)

            # Check if the AllPools collection already has this pool
            pool_helper = AllPools(self.chain_id).get(pool_address)
            if pool_helper:
                if TYPE_CHECKING:
                    assert isinstance(pool_helper, V3LiquidityPool)
                if pool_helper.factory == self._factory_address:
                    self._add_pool(pool_helper)
                    return pool_helper
                else:
                    self._untracked_pools.add(pool_address)
                    raise PoolNotAssociated(f"Pool {pool_address} is not associated with this DEX")

            if self._snapshot:
                v3liquiditypool_kwargs.update(
                    {
                        "tick_bitmap": self._snapshot.get_tick_bitmap(pool_address),
                        "tick_data": self._snapshot.get_tick_data(pool_address),
                    }
                )
            else:
                logger.info(
                    f"Initializing pool manager at address {self._factory_address} without liquidity snapshot"
                )

            # The pool is unknown, so build and add it
            try:
                pool_helper = V3LiquidityPool(
                    address=pool_address,
                    lens=self._lens,
                    silent=silent,
                    factory_address=self._factory_address,
                    factory_init_hash=self._factory_init_hash,
                    **v3liquiditypool_kwargs,
                    state_block=state_block,
                )
            except Exception as e:
                self._untracked_pools.add(pool_address)
                raise ManagerError(f"Could not build V3 pool {pool_address}: {e}") from e
            else:
                apply_liquidity_updates(pool_helper)
                self._add_pool(pool_helper)
                assert isinstance(
                    pool_helper, V3LiquidityPool
                ), f"{self} Attempted to return non-V3 pool {pool_helper}! {pool_address=}, {token_addresses=}, {pool_fee=}"
                return pool_helper

        if not (pool_address is None) ^ (token_addresses is None and pool_fee is None):
            raise ValueError("Insufficient arguments provided. Pass address OR tokens & fee")

        if v3liquiditypool_kwargs is None:
            v3liquiditypool_kwargs = dict()

        if pool_address is not None:
            # print(f"building V3 pool from address")
            if token_addresses is not None or pool_fee is not None:
                raise ValueError("Conflicting arguments provided. Pass address OR tokens+fee")
            pool_address = to_checksum_address(pool_address)
        elif token_addresses is not None and pool_fee is not None:
            # print(f"building V3 pool from address and fee")
            if len(token_addresses) != 2:
                raise ValueError(f"Expected two tokens, found {len(token_addresses)}")
            try:
                erc20token_helpers: List[Erc20Token] = [
                    self._token_manager.get_erc20token(
                        address=token_address,
                        silent=silent,
                    )
                    for token_address in token_addresses
                ]
            except Exception:
                raise ManagerError("Could not build Erc20Token helper")

            # dictionary key pair is sorted by address
            erc20token_helpers = sorted(erc20token_helpers)
            tokens_key: Tuple[ChecksumAddress, ChecksumAddress] = (
                erc20token_helpers[0].address,
                erc20token_helpers[1].address,
            )

            pool_address = generate_v3_pool_address(
                token_addresses=tokens_key,
                fee=pool_fee,
                factory_address=self._factory_address,
                init_hash=self._factory_init_hash,
            )
        else:
            raise ValueError("THIS BLOCK SHOULD BE UNREACHABLE")

        if pool_address in self._untracked_pools:
            raise PoolNotAssociated(
                f"Pool address {pool_address} not associated with factory {self._factory_address}"
            )

        try:
            pool_helper = self._tracked_pools[pool_address]
        except KeyError:
            pool_helper = find_or_build(pool_address, state_block)

        if TYPE_CHECKING:
            assert isinstance(pool_helper, V3LiquidityPool)

        assert isinstance(
            pool_helper, V3LiquidityPool
        ), f"{self} Attempted to return non-V3 pool {pool_helper}! {pool_address=}, {token_addresses=}, {pool_fee=}"
        return pool_helper



# ruff: noqa: F401


from . import exceptions, uniswap
from .arbitrage.arbitrage_dataclasses import ArbitrageCalculationResult
from .arbitrage.flash_borrow_to_lp_swap import FlashBorrowToLpSwap
from .arbitrage.flash_borrow_to_lp_swap_new import FlashBorrowToLpSwapNew
from .arbitrage.flash_borrow_to_lp_swap_with_future import FlashBorrowToLpSwapWithFuture
from .arbitrage.flash_borrow_to_router_swap import FlashBorrowToRouterSwap
from .arbitrage.uniswap_curve_cycle import UniswapCurveCycle
from .arbitrage.uniswap_lp_cycle import UniswapLpCycle
from .chainlink import ChainlinkPriceContract
from .config import get_web3, set_web3
from .curve.curve_stableswap_liquidity_pool import CurveStableswapPool
from .erc20_token import Erc20Token
from .fork.anvil_fork import AnvilFork
from .functions import next_base_fee
from .logging import logger
from .manager.token_manager import Erc20TokenHelperManager
from .registry.all_pools import AllPools
from .registry.all_tokens import AllTokens
from .transaction.uniswap_transaction import UniswapTransaction
from .uniswap.managers import (
    UniswapV2LiquidityPoolManager,
    UniswapV3LiquidityPoolManager,
)
from .uniswap.v2_liquidity_pool import CamelotLiquidityPool, LiquidityPool
from .uniswap.v3_dataclasses import (
    UniswapV3BitmapAtWord,
    UniswapV3LiquidityAtTick,
    UniswapV3LiquidityEvent,
    UniswapV3PoolExternalUpdate,
    UniswapV3PoolSimulationResult,
    UniswapV3PoolState,
)
from .uniswap.v3_liquidity_pool import V3LiquidityPool
from .uniswap.v3_snapshot import UniswapV3LiquiditySnapshot
from .uniswap.v3_tick_lens import TickLens



from typing import Any, Dict

from eth_typing import ChecksumAddress
from eth_utils.address import to_checksum_address


FACTORY_ADDRESSES: Dict[
    int,  # Chain ID
    Dict[
        ChecksumAddress,  # Factory address
        Dict[str, Any],
    ],
] = {
    1: {
        # Uniswap (V2)
        to_checksum_address("0x5C69bEe701ef814a2B6a3EDD4B1652CB9cc5aA6f"): {
            "init_hash": "0x96e8ac4277198ff8b6f785478aa9a39f403cb768dd02cbee326c3e7da348845f"
        },
        # Uniswap (V3)
        to_checksum_address("0x1F98431c8aD98523631AE4a59f267346ea31F984"): {
            "init_hash": "0xe34f199b19b2b4f47f68442619d555527d244f78a3297ea89325f843f87b8b54"
        },
        # Sushiswap (V2)
        to_checksum_address("0xC0AEe478e3658e2610c5F7A4A2E1777cE9e4f2Ac"): {
            "init_hash": "0xe18a34eb0e04b04f7a0ac29a6e80748dca96319b42c54d679cb821dca90c6303"
        },
        # Sushiswap (V3)
        to_checksum_address("0xbACEB8eC6b9355Dfc0269C18bac9d6E2Bdc29C4F"): {
            "init_hash": "0xe34f199b19b2b4f47f68442619d555527d244f78a3297ea89325f843f87b8b54"
        },
    },
    42161: {
        # Uniswap (V3)
        to_checksum_address("0x1F98431c8aD98523631AE4a59f267346ea31F984"): {
            "init_hash": "0xe34f199b19b2b4f47f68442619d555527d244f78a3297ea89325f843f87b8b54"
        },
        # Sushiswap (V2)
        to_checksum_address("0xc35DADB65012eC5796536bD9864eD8773aBc74C4"): {
            "init_hash": "0xe18a34eb0e04b04f7a0ac29a6e80748dca96319b42c54d679cb821dca90c6303"
        },
        # Sushiswap (V3)
        to_checksum_address("0x1af415a1EbA07a4986a52B6f2e7dE7003D82231e"): {
            "init_hash": "0xe34f199b19b2b4f47f68442619d555527d244f78a3297ea89325f843f87b8b54"
        },
    },
}


TICKLENS_ADDRESSES: Dict[
    int,  # Chain ID
    Dict[
        ChecksumAddress,  # Factory address
        ChecksumAddress,  # TickLens address
    ],
] = {
    # Ethereum Mainnet
    1: {
        # Uniswap V3
        # ref: https://docs.uniswap.org/contracts/v3/reference/deployments
        to_checksum_address("0x1F98431c8aD98523631AE4a59f267346ea31F984"): to_checksum_address(
            "0xbfd8137f7d1516D3ea5cA83523914859ec47F573"
        ),
        # Sushiswap V3
        # ref: https://docs.sushi.com/docs/Products/V3%20AMM/Periphery/Deployment%20Addresses
        to_checksum_address("0xbACEB8eC6b9355Dfc0269C18bac9d6E2Bdc29C4F"): to_checksum_address(
            "0xFB70AD5a200d784E7901230E6875d91d5Fa6B68c"
        ),
    },
    # Arbitrum
    42161: {
        # Uniswap V3
        # ref: https://docs.uniswap.org/contracts/v3/reference/deployments
        to_checksum_address("0x1F98431c8aD98523631AE4a59f267346ea31F984"): to_checksum_address(
            "0xbfd8137f7d1516D3ea5cA83523914859ec47F573"
        ),
        # Sushiswap V3
        # ref: https://docs.sushi.com/docs/Products/V3%20AMM/Periphery/Deployment%20Addresses
        to_checksum_address("0x1af415a1EbA07a4986a52B6f2e7dE7003D82231e"): to_checksum_address(
            "0x8516944E89f296eb6473d79aED1Ba12088016c9e"
        ),
    },
}



from degenbot import ChainlinkPriceContract, set_web3
from eth_utils import to_checksum_address


def test_chainlink_feed(ethereum_full_node_web3):
    set_web3(ethereum_full_node_web3)

    # Load WETH price feed
    # ref: https://data.chain.link/ethereum/mainnet/crypto-usd/eth-usd
    weth_price_feed = ChainlinkPriceContract(
        to_checksum_address("0x5f4ec3df9cbd43714fe2740f5e3616155c5b8419")
    )
    assert isinstance(weth_price_feed.update_price(), float)



import dataclasses
from typing import Any, List, Tuple

from ..erc20_token import Erc20Token


@dataclasses.dataclass(slots=True, frozen=True)
class ArbitrageCalculationResult:
    id: str
    input_token: Erc20Token
    profit_token: Erc20Token
    input_amount: int
    profit_amount: int
    swap_amounts: List[Any]


@dataclasses.dataclass(slots=True, frozen=True)
class CurveStableSwapPoolSwapAmounts:
    token_in: Erc20Token
    token_in_index: int
    token_out: Erc20Token
    token_out_index: int
    amount_in: int
    min_amount_out: int
    underlying: bool


@dataclasses.dataclass(slots=True, frozen=True)
class UniswapPoolSwapVector:
    token_in: Erc20Token
    token_out: Erc20Token
    zero_for_one: bool


@dataclasses.dataclass(slots=True, frozen=True)
class UniswapV2PoolSwapAmounts:
    amounts: Tuple[int, int]


@dataclasses.dataclass(slots=True, frozen=True)
class UniswapV3PoolSwapAmounts:
    amount_specified: int
    zero_for_one: bool
    sqrt_price_limit_x96: int


@dataclasses.dataclass(slots=True, frozen=True)
class CurveStableSwapPoolVector:
    token_in: Erc20Token
    token_out: Erc20Token



import dataclasses
from typing import TYPE_CHECKING, Any, Dict, Tuple

from ..baseclasses import BasePoolState, UniswapSimulationResult

if TYPE_CHECKING:
    from .v3_liquidity_pool import V3LiquidityPool


@dataclasses.dataclass(slots=True)
class UniswapV3BitmapAtWord:
    bitmap: int = 0
    block: int | None = dataclasses.field(compare=False, default=None)

    def to_dict(self) -> Dict[str, Any]:
        return dataclasses.asdict(self)


@dataclasses.dataclass(slots=True)
class UniswapV3LiquidityAtTick:
    liquidityNet: int = 0
    liquidityGross: int = 0
    block: int | None = dataclasses.field(compare=False, default=None)

    def to_dict(self) -> Dict[str, Any]:
        return dataclasses.asdict(self)


@dataclasses.dataclass(slots=True)
class UniswapV3LiquidityEvent:
    block_number: int
    liquidity: int
    tick_lower: int
    tick_upper: int
    tx_index: int


@dataclasses.dataclass(slots=True, eq=False)
class UniswapV3PoolExternalUpdate:
    block_number: int = dataclasses.field(compare=False)
    liquidity: int | None = None
    sqrt_price_x96: int | None = None
    tick: int | None = None
    liquidity_change: Tuple[
        int,  # Liquidity
        int,  # TickLower
        int,  # TickUpper
    ] | None = None
    tx: str | None = dataclasses.field(compare=False, default=None)


@dataclasses.dataclass(slots=True, frozen=True)
class UniswapV3PoolState(BasePoolState):
    pool: "V3LiquidityPool"
    liquidity: int
    sqrt_price_x96: int
    tick: int
    tick_bitmap: Dict[int, UniswapV3BitmapAtWord] | None = dataclasses.field(default=None)
    tick_data: Dict[int, UniswapV3LiquidityAtTick] | None = dataclasses.field(default=None)

    def copy(self) -> "UniswapV3PoolState":
        return UniswapV3PoolState(
            pool=self.pool,
            liquidity=self.liquidity,
            sqrt_price_x96=self.sqrt_price_x96,
            tick=self.tick,
            tick_bitmap=self.tick_bitmap.copy() if self.tick_bitmap is not None else None,
            tick_data=self.tick_data.copy() if self.tick_data is not None else None,
        )


@dataclasses.dataclass(slots=True, frozen=True)
class UniswapV3PoolSimulationResult(UniswapSimulationResult):
    current_state: UniswapV3PoolState = dataclasses.field(compare=False)
    future_state: UniswapV3PoolState = dataclasses.field(compare=False)



import logging

"""
Create a shared logger instance for this module. Set default level to INFO and send to stdout with StreamHandler
"""

logger = logging.getLogger("degenbot")
logger.propagate = False
logger.setLevel(logging.INFO)
logger.addHandler(logging.StreamHandler())



from web3 import Web3

from .exceptions import DegenbotError
from .logging import logger

_web3: Web3


def get_web3() -> Web3:
    try:
        return _web3
    except NameError:
        raise DegenbotError("A Web3 instance has not been provided.") from None


def set_web3(w3: Web3) -> None:
    if w3.is_connected() is False:
        raise DegenbotError("Web3 object is not connected.")

    logger.info(f"Connected to Web3 provider {w3.provider}")

    global _web3
    _web3 = w3



import os
import shutil
import socket
import subprocess
from typing import Any, Dict, Iterable, List, Tuple
from eth_typing import HexAddress
from hexbytes import HexBytes

import ujson
from eth_utils.address import to_checksum_address
from web3 import IPCProvider, Web3
from web3.types import Middleware

from ..constants import MAX_UINT256

_SOCKET_READ_BUFFER_SIZE = 4096  # https://docs.python.org/3/library/socket.html#socket.socket.recv


class AnvilFork:
    """
    Launch an Anvil fork via subprocess and provide various methods for
    interacting with it via JSON-RPC over the built-in IPC socket.
    """

    def __init__(
        self,
        fork_url: str,
        fork_block: int | None = None,
        hardfork: str = "latest",
        gas_limit: int = 30_000_000,
        port: int | None = None,
        chain_id: int | None = None,
        base_fee: int | None = None,
        ipc_path: str | None = None,
        mnemonic: str = (
            # Default mnemonic used by Brownie for Ganache forks
            "patient rude simple dog close planet oval animal hunt sketch suspect slim"
        ),
        coinbase: HexAddress | None = None,
        middlewares: List[Tuple[Middleware, int]] | None = None,
        balance_overrides: Iterable[Tuple[HexAddress, int]] | None = None,
        bytecode_overrides: Iterable[Tuple[HexAddress, bytes]] | None = None,
    ):
        if shutil.which("anvil") is None:  # pragma: no cover
            raise Exception("Anvil is not installed or not accessible in the current path.")

        if not port:
            with socket.socket() as sock:
                sock.bind(("", 0))
                _, port = sock.getsockname()
        self.port = port

        if not ipc_path:
            ipc_path = f"/tmp/anvil-{self.port}.ipc"

        command = []
        command.append("anvil")
        command.append("--silent")
        command.append("--auto-impersonate")
        command.append("--no-rate-limit")
        command.append(f"--fork-url={fork_url}")
        command.append(f"--hardfork={hardfork}")
        command.append(f"--gas-limit={gas_limit}")
        command.append(f"--port={self.port}")
        command.append(f"--ipc={ipc_path}")
        command.append(f"--mnemonic={mnemonic}")
        if fork_block:
            command.append(f"--fork-block-number={fork_block}")
        if chain_id:
            command.append(f"--chain-id={chain_id}")
        if base_fee:
            command.append(f"--base-fee={base_fee}")

        self._process = subprocess.Popen(command)
        self.fork_url = fork_url
        self.http_url = f"http://localhost:{self.port}"
        self.ws_url = f"ws://localhost:{self.port}"
        self.ipc_path = ipc_path
        self.w3 = Web3(IPCProvider(ipc_path))

        if middlewares is not None:
            for middleware, layer in middlewares:
                self.w3.middleware_onion.inject(middleware, layer=layer)

        while self.w3.is_connected() is False:
            continue

        self.socket = socket.socket(socket.AF_UNIX)
        self.socket.connect(self.ipc_path)

        self.block_number = fork_block if fork_block is not None else self.w3.eth.get_block_number()
        self.base_fee = (
            base_fee
            if base_fee is not None
            else self.w3.eth.get_block(self.block_number)["baseFeePerGas"]
        )
        self.base_fee_next: int | None = None
        self.chain_id: int = chain_id if chain_id is not None else self.w3.eth.chain_id

        if balance_overrides is not None:
            for account, balance in balance_overrides:
                self.set_balance(account, balance)

        if bytecode_overrides is not None:
            for account, bytecode in bytecode_overrides:
                self.set_code(account, bytecode)

        if coinbase is not None:
            self.set_coinbase(coinbase)

    def __del__(self) -> None:
        self._process.terminate()
        self._process.wait()
        if os.path.exists(self.ipc_path):
            os.remove(self.ipc_path)

    def _send_request(self, method: str, params: List[Any] | None = None) -> None:
        """
        Send a JSON-formatted request through the socket.
        """
        if params is None:
            params = []

        self.socket.sendall(
            bytes(
                ujson.dumps(
                    {
                        "jsonrpc": "2.0",
                        "id": None,
                        "method": method,
                        "params": params,
                    }
                ),
                encoding="utf-8",
            ),
        )

    def _get_response(self) -> Any:
        """
        Read the response payload from socket and return the JSON-decoded result.
        """
        raw_response = b""
        while True:
            raw_response += self.socket.recv(_SOCKET_READ_BUFFER_SIZE).rstrip()
            if raw_response.endswith((b"]", b"}")):
                # Valid JSON RPC responses will end in ] or }
                # ref: http://www.jsonrpc.org/specification
                response = ujson.loads(raw_response)
                if response.get("error"):
                    raise Exception(f"Error in response: {response}")
                return response["result"]

    def create_access_list(self, transaction: Dict[Any, Any]) -> Any:
        # Exclude transaction values that are irrelevant for the JSON-RPC method
        # ref: https://docs.infura.io/networks/ethereum/json-rpc-methods/eth_createaccesslist
        keys_to_drop = ("gasPrice", "maxFeePerGas", "maxPriorityFeePerGas", "gas", "chainId")
        sanitized_tx = {k: v for k, v in transaction.items() if k not in keys_to_drop}

        # Apply int->hex conversion to some transaction values
        # ref: https://docs.infura.io/networks/ethereum/json-rpc-methods/eth_createaccesslist
        keys_to_hexify = ("value", "nonce")
        for key in keys_to_hexify:
            if key in sanitized_tx:
                if isinstance(sanitized_tx[key], int):
                    sanitized_tx[key] = hex(sanitized_tx[key])

        self._send_request(
            method="eth_createAccessList",
            params=[sanitized_tx],
        )
        response: Dict[Any, Any] = self._get_response()
        return response["accessList"]

    def reset(
        self,
        fork_url: str | None = None,
        block_number: int | None = None,
    ) -> None:
        forking_params: Dict[str, Any] = {
            "jsonRpcUrl": fork_url if fork_url is not None else self.fork_url,
            "blockNumber": block_number if block_number is not None else self.block_number,
        }

        self._send_request(
            method="anvil_reset",
            params=[{"forking": forking_params}],
        )
        self._get_response()

        if block_number:
            self.block_number = block_number
        if fork_url:
            self.fork_url = fork_url

    def mine(self) -> None:
        self._send_request(method="evm_mine")
        self._get_response()

    def return_to_snapshot(self, id: int) -> bool:
        if id < 0:
            raise ValueError("ID cannot be negative")
        self._send_request(
            method="evm_revert",
            params=[id],
        )
        return bool(self._get_response())

    def set_balance(self, address: str, balance: int) -> None:
        if not (0 <= balance <= MAX_UINT256):
            raise ValueError("Invalid balance, must be within range: 0 <= balance <= 2**256 - 1")
        self._send_request(
            method="anvil_setBalance",
            params=[
                to_checksum_address(address),
                hex(balance),
            ],
        )
        self._get_response()

    def set_coinbase(self, address: str) -> None:
        self._send_request(
            method="anvil_setCoinbase",
            params=[HexBytes(address).hex()],
        )
        self._get_response()

    def set_code(self, address: str, bytecode: bytes) -> None:
        self._send_request(
            method="anvil_setCode",
            params=[
                HexBytes(address).hex(),
                HexBytes(bytecode).hex(),
            ],
        )
        self._get_response()

    def set_next_base_fee(self, fee: int) -> None:
        if not (0 <= fee <= MAX_UINT256):
            raise ValueError("Fee outside valid range 0 <= fee <= 2**256-1")
        self._send_request(
            method="anvil_setNextBlockBaseFeePerGas",
            params=[hex(fee)],
        )
        self._get_response()
        self.base_fee_next = fee

    def set_snapshot(self) -> int:
        self._send_request(
            method="evm_snapshot",
        )
        return int(self._get_response(), 16)



from typing import Any, Dict, Tuple


class MultiLiquidityPool:
    def __init__(self, *args: Tuple[Any], **kwargs: Dict[Any, Any]) -> None:  # pragma: no cover
        raise DeprecationWarning("This class has been deprecated in favor of UniswapLpCycle.")



from typing import Any

from degenbot.baseclasses import BaseLiquidityPool
from hexbytes import HexBytes


class FakePool(BaseLiquidityPool):
    # Override abstract methods
    def subscribe(self, subscriber: Any):
        return

    def get_arbitrage_helpers(self):
        return


def test_address_comparisons():
    P1_ADDRESS = "0x0a1b2c3d"
    P2_ADDRESS = "0x3d2c1b0a"

    p1 = FakePool()
    p2 = FakePool()

    p1.address = P1_ADDRESS
    p2.address = P2_ADDRESS

    assert p1 != p2
    assert p1 < p2
    assert p1 < P2_ADDRESS
    assert p1 < HexBytes(P2_ADDRESS)

    assert p2 > p1
    assert p2 > P1_ADDRESS
    assert p2 > HexBytes(P1_ADDRESS)

    assert p1 == P1_ADDRESS
    assert p2 == P2_ADDRESS

    assert p1 == P1_ADDRESS.lower()
    assert p1 == P1_ADDRESS.upper()
    assert p1 == HexBytes(P1_ADDRESS)

    assert p2 == P2_ADDRESS.lower()
    assert p2 == P2_ADDRESS.upper()
    assert p2 == HexBytes(P2_ADDRESS)



from web3.types import BlockIdentifier

from . import config
from .constants import MAX_UINT256


def get_number_for_block_identifier(identifier: BlockIdentifier | None) -> int:
    match identifier:
        case None:
            return config.get_web3().eth.get_block_number()
        case int() if 1 <= identifier <= MAX_UINT256:
            return identifier
        case bytes():
            return int.from_bytes(identifier, byteorder="big")
        case str() if isinstance(identifier, str) and identifier[:2] == "0x" and len(
            identifier
        ) == 66:
            return int(identifier, 16)
        case "latest" | "earliest" | "pending" | "safe" | "finalized":
            # These tags vary with each new block, so translate to a fixed block number
            return config.get_web3().eth.get_block(identifier)["number"]
        case _:
            raise ValueError(f"Invalid block identifier {identifier!r}")


def next_base_fee(
    parent_base_fee: int,
    parent_gas_used: int,
    parent_gas_limit: int,
    min_base_fee: int | None = None,
    base_fee_max_change_denominator: int = 8,  # limits the maximum base fee increase per block to 1/8 (12.5%)
    elasticity_multiplier: int = 2,
) -> int:
    """
    Calculate next base fee for an EIP-1559 compatible blockchain. The
    formula is taken from the example code in the EIP-1559 proposal (ref:
    https://eips.ethereum.org/EIPS/eip-1559).

    The default values for `base_fee_max_change_denominator` and
    `elasticity_multiplier` are taken from EIP-1559.

    Enforces `min_base_fee` if provided.
    """

    last_gas_target = parent_gas_limit // elasticity_multiplier

    if parent_gas_used == last_gas_target:
        next_base_fee = parent_base_fee
    elif parent_gas_used > last_gas_target:
        gas_used_delta = parent_gas_used - last_gas_target
        base_fee_delta = max(
            parent_base_fee * gas_used_delta // last_gas_target // base_fee_max_change_denominator,
            1,
        )
        next_base_fee = parent_base_fee + base_fee_delta
    else:
        gas_used_delta = last_gas_target - parent_gas_used
        base_fee_delta = (
            parent_base_fee * gas_used_delta // last_gas_target // base_fee_max_change_denominator
        )
        next_base_fee = parent_base_fee - base_fee_delta

    return max(min_base_fee, next_base_fee) if min_base_fee else next_base_fee
